import os
import re
from collections import defaultdict

from .basic_archive import BasicArchive, BASIC_ENTITY_TYPES, InterfaceError, ArchiveError
from .lc_archive import LcArchive, LC_ENTITY_TYPES
from ..implementations import BasicImplementation, IndexImplementation


class AbstractIndex(object):
    def make_interface(self, iface):
        if iface == 'index':
            return IndexImplementation(self)
        elif iface == 'basic':
            return BasicImplementation(self)
        else:
            raise InterfaceError('Unable to create interface %s' % iface)


class BasicIndex(AbstractIndex, BasicArchive):
    _ns_uuid_required = None

    def serialize(self, characterizations=False, values=False, domesticate=False):
        return super(BasicIndex, self).serialize(characterizations=False, values=False, domesticate=domesticate)

    def write_to_file(self, filename, gzip=True, **kwargs):
        """
        created to make gzip=True by default for index archives
        :param filename:
        :param gzip:
        :param kwargs:
        :return:
        """
        return super(BasicIndex, self).write_to_file(filename, gzip=gzip, **kwargs)


class LcIndex(AbstractIndex, LcArchive):
    _ns_uuid_required = None

    def serialize(self, exchanges=False, characterizations=False, values=False, domesticate=False):
        return super(LcIndex, self).serialize(exchanges=False, characterizations=False, values=False,
                                              domesticate=domesticate)

    def write_to_file(self, filename, gzip=True, **kwargs):
        """
        created to make gzip=True by default for index archives
        :param filename:
        :param gzip:
        :param kwargs:
        :return:
        """
        return super(LcIndex, self).write_to_file(filename, gzip=gzip, **kwargs)


def index_archive(archive, source, ref=None, signifier='index', force=False, save=True):
    if source is None:
        if save is True:
            raise AssertionError('Source is required')
    else:
        if not bool(re.search('\.gz$', source)):
            source += '.gz'
        if source == archive.source:
            raise ArchiveError('Index must have a different source from original archive')
        if save and os.path.exists(source):
            if force:
                print('File exists: %s. Overwriting..' % source)
            else:
                raise FileExistsError(source)

    if ref is None or ref == archive.ref:
        ref = archive.construct_new_ref(signifier=signifier)

    if isinstance(archive, LcArchive):
        index = LcIndex(source, ref=ref, static=True, **archive.init_args)
        types = LC_ENTITY_TYPES
    else:
        index = BasicIndex(source, ref=ref, static=True, **archive.init_args)
        types = BASIC_ENTITY_TYPES
    for t in types:
        for e in archive.entities_by_type(t):
            index.add(e)

    # import original archives list of names
    names = defaultdict(list)
    for s, r in archive.names.items():
        names[r].append(s)

    index.load_from_dict({'catalogNames': names})

    if save:
        print('Saving index to file %s' % source)
        index.write_to_file(source, gzip=True)
    return index



from antelope.models import ResponseModel, OriginMeta
from .rest_client import RestClient


class XdbRequester(RestClient):
    """
    A RestClient that encapsulates translating the HTTP responses to Pydantic models.  A token is structurally required
    but is still None default (will just give 401 unauthorized).  On initialization, queries API_root/origins to
    determine the set of origins the query knows.

    ref is optional and if supplied, the origin will be automatically prefixed to every query. Escape this behavior
    by using origin_x routes to supply origin explicitly.
    (if None, a client will be expected to specify origin in every query

    {{What *is* the difference between .origin and .ref? anyway? why the absurd machinations around catalog_names?
       ans: something (perhaps superstitious) about provenance- it's about mapping the ref to a 'physical' source }}

    pydantic operations
    use specified ref as origin
     -get_one
     -get_many
     -post_return_one
     -post_return_many
    user supplies origin as a positional parameter
     -origin_get_one
     -origin_get_many
     -origin_post_return_one
     -origin_post_return_many
    """
    def __init__(self, api_root, ref=None, token=None, quiet=False, **kwargs):
        super(XdbRequester, self).__init__(api_root, token=token, quiet=quiet, **kwargs)
        # retrieve origins authorized by this token
        self._origins = sorted((OriginMeta(**k) for k in self._get_endpoint('origins')),
                               key=lambda x: x.origin)
        self._org = ref  # '/'.join([api_root, origin])  # we prepend the API_ROOT now in the parent class

    @property
    def origin(self):
        return self._org

    @property
    def origins(self):
        """
        generates OriginMeta objects obtained by the server when the token was first authenticated
        Returns OriginMeta data-- this should probably include config information !
        :return:
        """
        for org in self._origins:
            yield org

    @property
    def is_lcia_engine(self):
        return any(k.is_lcia_engine for k in self._origins)

    def get_raw(self, *args, **kwargs):
        return self._get_endpoint(self._org, *args, **kwargs)

    def get_one(self, model, *args, **kwargs):
        if issubclass(model, ResponseModel):
            return model(**self._get_endpoint(self._org, *args, **kwargs))
        else:
            return model(self._get_endpoint(self._org, *args, **kwargs))

    def get_many(self, model, *args, **kwargs):
        if issubclass(model, ResponseModel):
            return [model(**k) for k in self._get_endpoint(self._org, *args, **kwargs)]
        else:
            return [model(k) for k in self._get_endpoint(self._org, *args, **kwargs)]

    def origin_get_one(self, model, *args, **kwargs):
        if issubclass(model, ResponseModel):
            return model(**self._get_endpoint(*args, **kwargs))
        else:
            return model(self._get_endpoint(*args, **kwargs))

    def origin_get_many(self, model, *args, **kwargs):
        if issubclass(model, ResponseModel):
            return [model(**k) for k in self._get_endpoint(*args, **kwargs)]
        else:
            return [model(k) for k in self._get_endpoint(*args, **kwargs)]

    def post_return_one(self, postdata, model, *args, **kwargs):
        if issubclass(model, ResponseModel):
            return model(**self._post(postdata, self._org, *args, **kwargs))
        else:
            return model(self._post(postdata, self._org, *args, **kwargs))

    def origin_post_return_one(self, postdata, model, *args, **kwargs):
        if issubclass(model, ResponseModel):
            return model(**self._post(postdata, *args, **kwargs))
        else:
            return model(self._post(postdata, *args, **kwargs))

    def post_return_many(self, postdata, model, *args, **kwargs):
        if issubclass(model, ResponseModel):
            return [model(**k) for k in self._post(postdata, self._org, *args, **kwargs)]
        else:
            return [model(k) for k in self._post(postdata, self._org, *args, **kwargs)]

    def origin_post_return_many(self, origin, postdata, model, *args, **kwargs):
        if issubclass(model, ResponseModel):
            return [model(**k) for k in self._post(postdata, origin, *args, **kwargs)]
        else:
            return [model(k) for k in self._post(postdata, origin, *args, **kwargs)]



"""
supplies an Antelope archive that contains the ref data.
"""
import csv

from antelope_core.archives import BasicArchive
from antelope_core.entities import LcFlow, LcQuantity, LcUnit
from antelope import QuantityInterface, EntityNotFound
from antelope_core.implementations import BasicImplementation
from antelope_core.characterizations import Characterization, DuplicateCharacterizationError
from antelope_core.lcia_engine import LciaEngine

from antelope_core.entities import MetaQuantityUnit

from .olca_accessor import OlcaAccessorV2
import logging


class OlcaRefQuantityImplementation(BasicImplementation, QuantityInterface):
    def get_canonical(self, quantity, **kwargs):
        return self._archive.tm.get_canonical(quantity)

    def factors(self, quantity, flowable=None, context=None, **kwargs):
        """
        always retrieving these from file when asked allows the files to be updated
        but maybe it would be smrter just to add them locally
        what would be smrtest would be to just store them in a dumb rdbms instead of from file. but think of the
        *immediate* use case.
        that is all for qdb

        :param quantity:
        :param flowable:
        :param context:
        :param kwargs:
        :return: generate Characterization objects
        """
        qq = self.get_canonical(quantity)
        cfs = dict()
        l_count = 0

        if flowable is not None:
            flowable = self._archive.tm.get_flowable(flowable)
        if context is not None:
            context = self._archive.tm[context]

        for factor in self._archive.factors_for_quantity(qq):
            flow = self._archive[factor['Flow']]
            fb = self._archive.tm.get_flowable(flow.name)

            if flowable is not None:
                if flowable != fb:
                    continue

            cx = self._archive.tm[flow.context]  # OLCA flows are all bound to a distinct context
            if context is not None:
                if cx != context:
                    continue

            rq = self.get_canonical(factor['Flow property'])
            key = (fb, factor['Flow property'], cx)
            if key not in cfs:
                cfs[key] = Characterization(fb, rq, qq, cx)

            rq_unit = factor['Flow unit']
            # [Factor] [Indicator] / [Flow unit]  * [convert to] / [convert from] = value [Indicator] / rq.unit
            value = float(factor['Factor']) * rq.convert(to=rq_unit)
            locale = factor['Location']
            if locale:
                l_count += 1
            else:
                locale = flow.locale

            try:
                cfs[key][locale] = value
            except DuplicateCharacterizationError:
                logging.warning('Duplicate Characterization for locale %s\nkey: %s' % (locale, key))

        if l_count:
            print('method %s: %d non-null Locations' % (qq.uuid, l_count))

        for v in cfs.values():
            yield v


class OpenLcaRefData(BasicArchive):
    """
    Modern OLCA data format should be easy to use. We are ignoring their unit conversions (which, fair, should be
    incorporated) ... categories is gone ... and they hve no flow properties to speak of (that I can find)
    so we simply do what we did below, adapted to the simpler data format.

    first flow properties
    then we replicate the meta-quantities load sequence from json-LD
    and we do on-demand CF synthesis via a custom Quantity implementation
    and we're done
    """

    _ns_uuid_required = None  # don't bother with this- our entities all have UUIDs already

    def make_interface(self, iface):
        if iface == 'quantity':
            return OlcaRefQuantityImplementation(self)
        return super(OpenLcaRefData, self).make_interface(iface)

    def __init__(self, source=None, **kwargs):
        super(OpenLcaRefData, self).__init__(source, ns_uuid=None, **kwargs)  # term_manager=LciaEngine(),
        self._index = dict()
        self._a = OlcaAccessorV2(source)
        self._olca_unitgroups = self._a.read_unit_groups()
        self._olca_locales = self._a.read_locales()
        # self.load_all()  for profiling reasons, keep this separate

    def _fetch(self, entity, **kwargs):
        """
        :param entity:
        :param kwargs:
        :return:
        """
        ent = self[entity]
        if ent is None:
            raise KeyError(entity)
        else:
            return ent

    def _load_flow_properties(self):
        for fp in self._a.get_refdata('flow_properties'):
            if self[fp['ID']] is None:
                ug = self._olca_unitgroups[fp['Unit group']]

                q = LcQuantity(fp['ID'], Name=fp['Name'], ReferenceUnit=LcUnit(ug.ref_unit),
                               entity_uuid=fp['ID'], Category=fp['Category'], Type=fp['Property type'],
                               UnitConversion=ug.conversion)
                self.add(q)

    def _load_flows(self):
        for flow in self._a.get_refdata('flows'):
            f_id = flow.pop('ID')
            if self[f_id] is None:
                ref_prop = flow.pop('Reference flow property')
                try:
                    fp = self.tm.get_canonical(ref_prop)
                except EntityNotFound:
                    print('Flow %s: reference property not found: %s' % (f_id, ref_prop))
                    continue
                cx = tuple(flow.pop('Category').split('/'))
                flow['CasNumber'] = flow.pop('CAS number')

                # check for locale
                name = flow['Name']
                maybe_loc = name.split(',')[-1].strip()
                if maybe_loc in self._olca_locales:
                    flow['locale'] = maybe_loc

                f = LcFlow(f_id, ReferenceQuantity=fp, entity_uuid=f_id, context=cx, **flow)
                self.add(f)

    def _load_impact_categories(self):
        """
        lcia_categories: ID,Name,Description,Category,Reference unit
         it appears "category" is misnamed here- it actually is FK to lcia_methods.Name
         "description" is non-null for only a few.
        :return:
        """
        for lcia in self._a.get_refdata('lcia_categories'):
            l_id = lcia.pop('ID')
            if self[l_id] is None:
                method = lcia.pop('Category')
                name = lcia.pop('Name')
                indicator = lcia.pop('Reference unit')
                u = LcUnit(indicator)
                q = LcQuantity(l_id, ReferenceUnit=u, Name=name,
                               Method=method,
                               Category=name,
                               Indicator=indicator, normSets=[], normalisationFactors=[], weightingFactors=[])
                self.add(q)

    def _load_nw(self):
        """
        lcia_method_nw_sets: LCIA method,NW set - ID,NW set - name,LCIA category,Nomalisation factor,Weighting factor,Weighting score unit
         Weighting score unit appears null for all records, so we ignore it

        :return:
        """
        for nw in self._a.get_refdata('lcia_method_nw_sets'):
            q = self[nw['LCIA category']]
            if q['Method'] != nw['LCIA method']:
                print('!! Method disagreement %s: %s vs %s' % (q.uuid, q['Method'], nw['LCIA method']))

            # get the data
            set_name = nw['NW set - name']

            if nw['Weighting score unit']:
                print('%s: Weighting score unit found for %s [%s]' % (q.uuid, set_name, nw['Weighting score unit']))

            try:
                norm = float(nw['Nomalisation factor'])  # spelling!!!!
            except ValueError:
                norm = None
            try:
                wgt = float(nw['Weighting factor'])
            except ValueError:
                wgt = None
            if norm is None and wgt is None:
                print('%s: No values for weighting set %s' % (q.uuid, set_name))
                continue

            # store the data
            if set_name in q['normSets']:
                ix = q['normSets'].index(set_name)
                # the lists should always be the same length, but let's just pad them up
                while len(q['normalisationFactors']) < ix:
                    q['normalisationFactors'].append(0)
                while len(q['weightingFactors']) < ix:
                    q['weightingFactors'].append(0)
                q['normalisationFactors'][ix] = norm
                q['weightingFactors'][ix] = wgt
            else:
                q['normSets'].append(set_name)
                q['normalisationFactors'].append(norm)
                q['weightingFactors'].append(wgt)

    def _load_impact_methods(self):
        """
        the key question here is HOW to manage the normalization information. We had an ad hoc strategy for OpenLCA
        where each INDICATOR was given three properties: normSets, normalisationFactors, and weightingFactors,
        all lists, where the ith element of each list was in correspondence with one another.

        Then when we moved to pydantic, we created a Normalizations object which stored them as dicts.

        For the time being, let's just do what we said before and replicate the OpenLCA JsonLD approach.

        lcia_methods: ID,Name,Description,Category
         description is LONG STRING. Category is not interesting.
        lcia_method_categories: LCIA method,LCIA category
        :return:
        """
        categories = self._a.collate_lcia_categories()

        for im in self._a.get_refdata('lcia_methods'):
            if self[im['ID']] is None:
                method = im['Name']
                desc = im['Description']
                cat = im['Category']
                qs = categories.get(method, [])
                m = LcQuantity(method, Name=method, ReferenceUnit=MetaQuantityUnit, Method=method, Description=desc,
                               entity_uuid=im['ID'], Category=cat,
                               ImpactCategories=qs)
                self.add(m)
                if self[im['ID']] is not m:
                    raise ValueError

    def _load_all(self, flows=True, **kwargs):
        self._load_flow_properties()
        self._load_impact_categories()
        self._load_nw()
        self._load_impact_methods()
        if flows:
            self._load_flows()

    def factors_for_quantity(self, q):
        for cf in self._a.factors_for_method(q):
            yield cf

    def factors_to_csv(self, quantity, filepath):
        """
        Utility to output processed CFs to a file for easy analysis, debug unit conversions, duplicate CFs, etc
        :param quantity:
        :param filepath:
        :return:
        """
        qq = self.tm.get_canonical(quantity)

        fields = ('quantity',
                  'ref_quantity',
                  'flowable',
                  'context',
                  'locale',
                  'value',
                  'indicator',
                  'ref_unit')

        count = 0
        with open(filepath, 'w') as fp:
            cw = csv.DictWriter(fp, fieldnames=fields)
            for factor in self.factors_for_quantity(qq):
                flow = self[factor['Flow']]
                rq = flow.reference_entity
                fb = self.tm.get_flowable(flow.name)
                cx = self.tm[flow.context]
                locale = factor['Location']
                if not locale:
                    locale = flow.locale

                value = float(factor['Factor']) * rq.convert(to=factor['Flow unit'])

                row = {'quantity': qq.uuid,
                       'ref_quantity': rq.uuid,
                       'flowable': str(fb),
                       'context': '; '.join(cx.as_list()),
                       'locale': locale,
                       'value': value,
                       'indicator': qq['Indicator'],
                       'ref_unit': rq.unit}

                cw.writerow(row)
                count += 1
        print('Wrote %d rows to %s' % (count, filepath))



# from collections import defaultdict
import json
from antelope import (IndexInterface, ExchangeInterface, QuantityInterface, BackgroundInterface, NoFactorsFound,
                      ItemNotFound)
from antelope import RxRef, EntityNotFound
from antelope.refs.base import NoUuid
from antelope.models import (OriginCount, Entity, FlowEntity, Exchange, ReferenceExchange, ReferenceValue,
                             UnallocatedExchange,
                             LciaResult as LciaResultModel, AllocatedExchange,
                             Characterization as CharacterizationModel,
                             ExchangeValues, DirectedFlow, FlowFactors)

from antelope_core.implementations import BasicImplementation, ConfigureImplementation
from antelope_core.lcia_results import LciaResult
from antelope_core.characterizations import QRResult
from .xdb_entities import XdbReferenceRequired


from requests.exceptions import HTTPError


class BadClientRequest(Exception):
    pass


class XdbConfigureImplementation(ConfigureImplementation):
    def apply_config(self, config, **kwargs):
        r_config = self._archive.r.get_raw('config')
        # no other config needs to be run for now
        config.update(r_config)
        super(XdbConfigureImplementation, self).apply_config(config, **kwargs)


class RemoteExchange(Exchange):
    @property
    def is_reference(self):
        return self.type == 'reference'


class RemoteExchangeValues(ExchangeValues):
    @property
    def is_reference(self):
        return self.type == 'reference'


def _ref(obj):
    """
    URL-ize input argument... add underscores as brackets if the ref contains slashes
    :param obj:
    :return:
    """
    if hasattr(obj, 'external_ref'):
        ref = str(obj.external_ref)
    else:
        ref = str(obj)
    if ref.find('/') >= 0:
        return '_/%s/_' % ref
    else:
        return ref


class XdbImplementation(BasicImplementation, IndexInterface, ExchangeInterface, QuantityInterface, BackgroundInterface):
    """
    The implementation is very thin, so pile everything into one class
    """
    def setup_bm(self, query):
        return True

    def check_bg(self, **kwargs):
        return True

    def get(self, external_ref, origin=None, **kwargs):
        return self._archive.retrieve_or_fetch_entity(external_ref, origin=origin, **kwargs)

    def get_reference(self, key):
        p = self.get(key)
        if p.entity_type == 'process':
            try:
                rs = p.ref.get(p.ref.reference_field)
                if rs is None:
                    raise XdbReferenceRequired
                return [RxRef(p, self._archive.get_or_make(r.flow), r.direction, comment=r.comment)
                        for r in rs]
            except XdbReferenceRequired:
                rs = self._archive.r.get_many(ReferenceValue, _ref(key), 'references')
                return [RxRef(p, self._archive.get_or_make(r.flow), r.direction, comment=r.comment, value=r.value)
                        for r in rs]
        elif p.entity_type == 'flow':
            ref_q = self._archive.r.get_one(Entity, _ref(key), 'reference')
            return self._archive.get_or_make(ref_q)
        elif p.entity_type == 'quantity':
            return self._archive.r.get_one(str, _ref(key), 'reference')
        else:
            raise TypeError(p.entity_type)

    def properties(self, external_ref, **kwargs):
        the_ref = self.get(external_ref, **kwargs)
        props = self._archive.r.get_one(dict, _ref(external_ref), 'properties')
        for k, v in props.items():
            the_ref[k] = v
            yield k

    def get_item(self, external_ref, item):
        try:
            return self._archive.r.get_raw(_ref(external_ref), 'doc', item)
        except HTTPError as e:
            if e.args[0] == 404:
                raise ItemNotFound(external_ref, item)
            raise e

    def get_uuid(self, external_ref):
        """
        Stopgap: don't support UUIDs
        :param external_ref:
        :return:
        """
        try:
            return self._archive.r.get_raw(_ref(external_ref), 'uuid')
        except HTTPError as e:
            if e.args[0] == 404:
                return NoUuid
            raise

    '''
    Index routes
    '''
    def count(self, entity_type, **kwargs):
        """
        Naturally the first route is problematic- because we allow incompletely-specified origins.
        We should sum them.
        :param entity_type:
        :param kwargs:
        :return:
        """
        return sum(k.count[entity_type] for k in self._archive.r.get_many(OriginCount, 'count'))

    def processes(self, **kwargs):
        llargs = {k.lower(): v for k, v in kwargs.items()}
        return [self._archive.get_or_make(k) for k in self._archive.r.get_many(Entity, 'processes', **llargs)]

    def flows(self, **kwargs):
        llargs = {k.lower(): v for k, v in kwargs.items()}
        return [self._archive.get_or_make(k) for k in self._archive.r.get_many(FlowEntity, 'flows', **llargs)]

    def quantities(self, **kwargs):
        llargs = {k.lower(): v for k, v in kwargs.items()}
        return [self._archive.get_or_make(k) for k in self._archive.r.get_many(Entity, 'quantities', **llargs)]

    def lcia(self, **kwargs):
        llargs = {k.lower(): v for k, v in kwargs.items()}
        return [self._archive.get_or_make(k) for k in self._archive.r.get_many(Entity, 'lcia', **llargs)]

    def lcia_methods(self, **kwargs):
        llargs = {k.lower(): v for k, v in kwargs.items()}
        return [self._archive.get_or_make(k) for k in self._archive.r.get_many(Entity, 'lcia_methods', **llargs)]

    def contexts(self, **kwargs):
        return self._archive.tm.contexts(**kwargs)

    def get_context(self, term, **kwargs):
        if isinstance(term, list) or isinstance(term, tuple):
            return self._archive.tm.get_context(term[-1])
        return self._archive.tm.get_context(term)

    def targets(self, flow, direction=None, **kwargs):
        return [self._archive.get_or_make(k) for k in self._archive.r.get_many(Entity, _ref(flow), 'targets')]

    '''
    Exchange routes
    '''
    def _resolve_ex(self, ex):
        # self.get_canonical(ex.flow.quantity_ref)  # wtf was this for
        ex.process = self._archive.get(ex.process)
        ex.flow = self._archive.get_or_make(FlowEntity.from_exchange_model(ex))  # must get turned into a ref with make_ref

        if ex.type == 'context':
            ex.termination = self.get_context(ex.context)
        elif ex.type == 'cutoff':
            ex.termination = None
        return ex

    def _resolve_exv(self, exv: ExchangeValues):
        exv = self._resolve_ex(exv)
        if 'None' in exv.values:
            exv.values[None] = exv.values.pop('None')
        return exv

    def exchanges(self, process, **kwargs):
        """
        Client code (process_ref.ProcessRef) already turns them into ExchangeRefs
        :param process:
        :param kwargs:
        :return:
        """
        return list(self._resolve_ex(ex) for ex in self._archive.r.get_many(RemoteExchange, _ref(process), 'exchanges'))

    def exchange_values(self, process, flow, direction=None, termination=None, reference=None, **kwargs):
        """

        :param process:
        :param flow:
        :param direction:
        :param termination:
        :param reference:
        :param kwargs:
        :return:
        """
        exvs = list(self._resolve_exv(exv) for exv in self._archive.r.get_many(RemoteExchangeValues, _ref(process),
                                                                               'exchanges', _ref(flow)))
        if direction:
            exvs = list(filter(lambda x: x.direction == direction, exvs))
        if reference:
            exvs = list(filter(lambda x: x.is_reference == reference, exvs))
        return exvs

    def inventory(self, node, ref_flow=None, scenario=None, **kwargs):
        """
        Client code (process_ref.ProcessRef) already turns them into ExchangeRefs
        :param node:
        :param ref_flow: if node is a process, optionally provide its reference flow
        :param scenario: if node is a fragment, optionally provide a scenario- as string or tuple
        :param kwargs:
        :return:
        """
        if ref_flow and scenario:
            raise BadClientRequest('cannot specify both ref_flow and scenario')
        if ref_flow:
            # process inventory
            return list(self._resolve_ex(ex)
                        for ex in self._archive.r.get_many(AllocatedExchange, _ref(node), _ref(ref_flow), 'inventory'))
        elif scenario:
            return list(self._resolve_ex(ex)
                        for ex in self._archive.r.get_many(AllocatedExchange, _ref(node), 'inventory',
                                                           scenario=scenario))
        else:
            return list(self._resolve_ex(ex) for ex in self._archive.r.get_many(UnallocatedExchange, _ref(node),
                                                                                'inventory'))

    def exchange_relation(self, process, ref_flow, exch_flow, direction, termination=None, **kwargs):
        if direction is None:
            if ref_flow is None:
                return self._archive.r.get_one(float, 'exchange_relation', _ref(process), _ref(exch_flow))
            else:
                return self._archive.r.get_one(float, 'exchange_relation', _ref(process), _ref(ref_flow),
                                               _ref(exch_flow))
        else:
            if ref_flow is None:
                return self._archive.r.get_one(float, 'exchange_relation', _ref(process), _ref(exch_flow), direction)
            else:
                return self._archive.r.get_one(float, 'exchange_relation', _ref(process), _ref(ref_flow),
                                               _ref(exch_flow), direction)

    def dependencies(self, process, ref_flow=None, **kwargs):
        if ref_flow:
            # process inventory
            return list(self._resolve_ex(ex) for ex in self._archive.r.get_many(AllocatedExchange, _ref(process),
                                                                                _ref(ref_flow), 'dependencies'))
        else:
            return list(self._resolve_ex(ex) for ex in self._archive.r.get_many(AllocatedExchange, _ref(process),
                                                                                'dependencies'))

    def consumers(self, process, ref_flow=None, **kwargs):
        """
        This returns reference exchanges for activities that consume the named reference exchange
        the puzzle here is how to generate the RxRefs-- I guess we should just do it here
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        if ref_flow:
            return list(self._resolve_ex(rx) for rx in self._archive.r.get_many(ReferenceExchange,
                                                                                _ref(process), _ref(ref_flow),
                                                                                'consumers'))
        else:
            return list(self._resolve_ex(rx) for rx in self._archive.r.get_many(ReferenceExchange,
                                                                                _ref(process),
                                                                                'consumers'))

    def emissions(self, process, ref_flow=None, **kwargs):
        if ref_flow:
            # process inventory
            return list(self._resolve_ex(ex) for ex in self._archive.r.get_many(AllocatedExchange, _ref(process),
                                                                                _ref(ref_flow), 'emissions'))
        else:
            return list(self._resolve_ex(ex) for ex in self._archive.r.get_many(AllocatedExchange, _ref(process),
                                                                                'emissions'))

    def cutoffs(self, process, ref_flow=None, **kwargs):
        if ref_flow:
            # process inventory
            return list(self._resolve_ex(ex) for ex in self._archive.r.get_many(AllocatedExchange, _ref(process),
                                                                                _ref(ref_flow), 'cutoffs'))
        else:
            return list(self._resolve_ex(ex) for ex in self._archive.r.get_many(AllocatedExchange, _ref(process),
                                                                                'cutoffs'))

    def lci(self, process, ref_flow=None, **kwargs):
        if ref_flow:
            # process inventory
            return list(self._resolve_ex(ex)
                        for ex in self._archive.r.get_many(AllocatedExchange, _ref(process), _ref(ref_flow), 'lci'))
        else:
            return list(self._resolve_ex(ex) for ex in self._archive.r.get_many(AllocatedExchange, _ref(process), 'lci'))

    def _to_exch_ref(self, x):
        """
        We can't resolve references from here! something is fucked
        :param x:
        :return:
        """
        pass

    def sys_lci(self, demand, **kwargs):
        dmd = [UnallocatedExchange.from_inv(x).dict() for x in demand]
        return list(self._resolve_ex(ex)  # DWR! THESE ARE NOT OPERATIONAL EXCHANGES YET!!!
                    for ex in self._archive.r.post_return_many(dmd, UnallocatedExchange, 'sys_lci', **kwargs))

    '''
    qdb routes
    '''
    def factors(self, quantity, flowable=None, context=None, **kwargs):
        """
        We need to construct operable characterizations with quantities that are recognized by the LciaEngine- in other
        words, with refs from our archive
        :param quantity:
        :param flowable:
        :param context: not implemented at the API
        :param kwargs:
        :return:
        """
        if flowable:
            facs = self._archive.r.get_many(CharacterizationModel, quantity, 'factors', flowable)
        else:
            facs = self._archive.r.get_many(CharacterizationModel, quantity, 'factors')
        return list(facs)

    def cf(self, flow, quantity, ref_quantity=None, context=None, locale='GLO', **kwargs):
        """
        We still want to retain the ability to ask the remote server for CFs, even if we may prefer to get that
        info locally for local flows
        :param flow:
        :param quantity:
        :param ref_quantity: NOT USED
        :param context:
        :param locale:
        :param kwargs:
        :return:
        """
        try:
            return self._archive.r.get_one(float, _ref(flow), 'cf', _ref(quantity), context=context, locale=locale,
                                           ref_quantity=ref_quantity)
        except HTTPError as e:
            if e.args[0] == 404:
                return 0.0
            else:
                raise

    def characterize(self, flowable, ref_quantity, query_quantity, value, context=None, location='GLO', **kwargs):
        if context:
            context = self.get_context(context).as_list()
        if not isinstance(value, dict):
            value = {location: value}
        cf = CharacterizationModel(origin=self.origin, flowable=flowable, ref_quantity=_ref(ref_quantity),
                                   query_quantity=_ref(query_quantity), context=context,
                                   value=value)
        return self._archive.r.post_return_one(cf.model_dump(), CharacterizationModel, 'characterize', **kwargs)

    def quantity_relation(self, flowable, ref_quantity, query_quantity, context, locale='GLO', **kwargs):
        """
        not yet implemented
        :param flowable:
        :param ref_quantity:
        :param query_quantity:
        :param context:
        :param locale:
        :param kwargs:
        :return:
        """
        print('quantity_relation not implemented! %s, %s, %s' % (flowable, _ref(ref_quantity), _ref(query_quantity)))
        raise NoFactorsFound

    @staticmethod
    def _result_from_exchanges(quantity, exch_map, res_m: LciaResultModel):
        """
        Constructs a detailed LCIA result using details provided by the backend server, populated with exchanges
        that we provided via POST.

        :param quantity:
        :param exch_map:
        :param res_m:
        :return:
        """
        res = LciaResult(quantity, scenario=res_m.scenario, scale=res_m.scale)
        for c in res_m.components:
            for d in c.details:
                key = (d.exchange.external_ref, tuple(d.exchange.context))
                try:
                    ex = exch_map[key]
                except KeyError:
                    print('missing key %s,%s' % key)
                    continue
                val = d.result / d.factor.value
                if val != ex.value:
                    print('%s: value mismatch %g vs %g' % (key, val, ex.value))
                cf = QRResult(d.factor.flowable, ex.flow.reference_entity, quantity, ex.termination,
                              d.factor.locale, d.factor.origin, d.factor.value)
                res.add_score(c.component, ex, cf)
        if len(res_m.summaries) > 0:
            print('Ignoring spurious summaries:')
            for s in res_m.summaries:
                print(s)
        # the results here shouldn't have any summaries
        # for s in res_m.summaries:
        #     res.add_summary(s.component, node, s.node_weight, s.unit_score)
        return res

    def get_factors(self, quantity, flow_specs, **kwargs):
    # def _result_from_model(self, process_ref, quantity, res_m: LciaResultModel):
        """

        :param quantity:
        :param flow_specs:
        :param kwargs:
        :return:
        """
        specs = [fs.dict() for fs in flow_specs]
        return self._archive.r.origin_post_return_many('qdb', specs, FlowFactors,
                                                       _ref(quantity), 'flow_specs')

    def do_lcia(self, quantity, inventory, locale='GLO', **kwargs):
        """

        :param quantity:
        :param inventory:
        :param locale:
        :param kwargs:
        :return:
        """
        exchanges = [UnallocatedExchange.from_inv(x).dict() for x in inventory]
        exch_map = {(x.flow.external_ref, x.term_ref): x for x in inventory}

        res = self._archive.r.origin_post_return_one('qdb', exchanges, LciaResultModel, _ref(quantity), 'do_lcia')
        return self._result_from_exchanges(quantity, exch_map, res)

    def bg_lcia(self, process, query_qty=None, ref_flow=None, **kwargs):
        """

        :param process:
        :param query_qty:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        try:
            if query_qty is None:
                if ref_flow:
                    ress = self._archive.r.get_many(LciaResultModel, _ref(process), _ref(ref_flow),
                                                    'lcia', **kwargs)
                else:
                    ress = self._archive.r.get_many(LciaResultModel, _ref(process),
                                                    'lcia', **kwargs)
            else:
                if ref_flow:
                    ress = self._archive.r.get_one(LciaResultModel, _ref(process), _ref(ref_flow),
                                                   'lcia', _ref(query_qty), **kwargs)
                else:
                    ress = self._archive.r.get_one(LciaResultModel, _ref(process),
                                                   'lcia', _ref(query_qty), **kwargs)
        except HTTPError as e:
            if e.args[0] == 404:
                content = json.loads(e.args[1])
                raise EntityNotFound(content['detail'])
            else:
                raise
        return ress

    def sys_lcia(self, process, query_qty, observed=None, ref_flow=None, **kwargs):
        """
        We want to override the interface implementation and send a simple request to the backend
        :param process:
        :param query_qty:
        :param observed: iterable of DirectedFlows
        :param ref_flow:
        :param kwargs: locale, quell_biogenic_co2
        :return:
        """
        obs_flows = ()
        try:
            if observed:
                obs_flows = [k.model_dump() for k in observed]
            if len(obs_flows) > 0:
                if ref_flow:
                    ress = self._archive.r.post_return_one(obs_flows, LciaResultModel, _ref(process), _ref(ref_flow),
                                                           'lcia', _ref(query_qty), **kwargs)
                else:
                    ress = self._archive.r.post_return_one(obs_flows, LciaResultModel, _ref(process),
                                                           'lcia', _ref(query_qty), **kwargs)
            else:
                if ref_flow:
                    ress = self._archive.r.get_one(LciaResultModel, _ref(process), _ref(ref_flow),
                                                   'lcia', _ref(query_qty), **kwargs)
                else:
                    ress = self._archive.r.get_one(LciaResultModel, _ref(process),
                                                   'lcia', _ref(query_qty), **kwargs)
        except HTTPError as e:
            if e.args[0] == 404:
                content = json.loads(e.args[1])
                raise EntityNotFound(content['detail'])
            else:
                raise
        return ress  # [self._result_from_model(process, query_qty, res) for res in ress]



import requests
from typing import Optional
from requests.exceptions import HTTPError
import json
from time import time
from pydantic import BaseModel
import getpass


class OAuthToken(BaseModel):
    token_type: str
    access_token: str
    message: Optional[str] = None
    expires_in: Optional[int] = None

    @property
    def auth(self):
        return '%s %s' % (self.token_type, self.access_token)


class NoCredentials(Exception):
    pass


class RestClient(object):
    """
    A REST client that uses pydantic models to interpret response data
    """

    auth_route = 'login'
    _token = None

    def _print(self, *args, cont=False):
        if not self._quiet:
            if cont:  # continue the same line
                print(*args, end=".. ")
            else:
                print(*args)

    def __init__(self, api_root, token=None, quiet=False, auth_route=None, save_credentials=True, verify=None):
        self._s = requests.Session()
        if verify:
            self._s.verify = verify
        self._s.headers['Accept'] = "application/json"
        self._quiet = quiet
        if auth_route:
            self.auth_route = auth_route
        if token:
            self.set_token(token)

        self._save = bool(save_credentials)
        self._creds = None

        while api_root[-1] == '/':
            api_root = api_root[:-1]  # strip trailing /

        self._api_root = api_root

    @property
    def saved(self):
        return self._save and bool(self._creds)

    def close(self):
        self._s.close()

    def set_token(self, token):
        """
        we accept:
        - a string with a plain token
        - a string with 'token_type token'
        - an OauthToken
        :param token:
        :return:
        """
        if isinstance(token, OAuthToken):
            self._token = token
            if token.message:
                print(token.message)
            if token.expires_in:
                print('token expires in %d s' % token.expires_in)
        elif isinstance(token, str):
            toks = token.split(' ')
            if len(toks) == 1:
                self._token = OAuthToken(token_type='bearer', access_token=token)
            elif len(toks) == 2:
                self._token = OAuthToken(token_type=toks[0], access_token=toks[1])
            else:
                raise ValueError('invalid token specification')
        else:
            raise ValueError('Invalid token type %s' % type(token))
        self._s.headers['Authorization'] = self._token.auth

    def _upd_save(self, save_credentials):
        """
        Updates the flag indicating whether we are saving credentials. None = do nothing. clears the cache if false.
        :param save_credentials:
        :return:
        """
        if save_credentials is None:
            return
        else:
            self._save = bool(save_credentials)
        if not self._save:
            self._creds = None

    def reauthenticate(self, save_credentials=None, **kwargs):
        """
        If we have saved credentials, pass them along. update the flag
        :param save_credentials:
        :param kwargs:
        :return:
        """
        if self._creds is None:
            raise NoCredentials
        data = self._creds
        data.update(kwargs)
        self._upd_save(save_credentials)
        self._post_credentials(data)

    def authenticate(self, username, password=None, save_credentials=None, **kwargs):
        """
        POSTs an OAuth2-compliant form to obtain a bearer token.
        Be sure to set the 'auth_route' property either in a subclass or manually (e.g. on init)
        :param username:
        :param password:
        :param save_credentials: whether to save credentials
        :param kwargs:
        :return:
        """
        if password is None:
            password = getpass.getpass('Enter password for user %s: ' % username)
        data = {
            "grant_type": "password",
            "username": username,
            "password": password,
        }
        data.update(kwargs)
        self._upd_save(save_credentials)
        self._post_credentials(data)

    def _post_credentials(self, data):
        """
        Post the credentials, and save them if the request was successful and if thje save flag is set
        :param data:
        :return:
        """
        if not self._save:
            self._creds = None
        self.set_token(self.post_return_one(data, OAuthToken, self.auth_route, form=True))
        if self._save:
            self._creds = data

    def _request(self, verb, route, **kwargs):
        """
        Returns JSON-translated content
        :param verb:
        :param route:
        :param kwargs:
        :return:
        """
        url = '/'.join([self._api_root, route])
        endp = {
            'GET': self._s.get,
            'PUT': self._s.put,
            'POST': self._s.post,
            'PATCH': self._s.patch,
            'DELETE': self._s.delete
        }[verb]
        self._print('%s %s' % (verb, url), cont=True)
        t = time()
        resp = endp(url, **kwargs)
        el = time() - t
        self._print('%d [%.2f sec]' % (resp.status_code, el))
        if resp.status_code >= 400:
            raise HTTPError(resp.status_code, resp.content)
        return json.loads(resp.content)

    def _get_endpoint(self, route, *args, **params):
        url = '/'.join(map(str, filter(None, [route, *args])))
        return self._request('GET', url, params=params)

    def get_raw(self, *args, **kwargs):
        return self._get_endpoint(*args, **kwargs)

    def get_one(self, model, *args, **kwargs):
        if issubclass(model, BaseModel):
            return model(**self._get_endpoint(*args, **kwargs))
        else:
            return model(self._get_endpoint(*args, **kwargs))

    def get_many(self, model, *args, **kwargs):
        if issubclass(model, BaseModel):
            return [model(**k) for k in self._get_endpoint(*args, **kwargs)]
        else:
            return [model(k) for k in self._get_endpoint(*args, **kwargs)]

    def _post(self, postdata, route, *args, form=False, **params):
        url = '/'.join(map(str, [route, *args]))
        if form:
            return self._request('POST', url, data=postdata, params=params)
        else:
            return self._request('POST', url, json=postdata, params=params)

    def post_return_one(self, postdata, model, *args, **kwargs):
        if issubclass(model, BaseModel):
            return model(**self._post(postdata, *args, **kwargs))
        else:
            return model(self._post(postdata, *args, **kwargs))

    def post_return_many(self, postdata, model, *args, **kwargs):
        if issubclass(model, BaseModel):
            return [model(**k) for k in self._post(postdata, *args, **kwargs)]
        else:
            return [model(k) for k in self._post(postdata, *args, **kwargs)]

    def put(self, putdata, model, *args, form=False, **kwargs):
        url = '/'.join(map(str, args))
        if form:
            response = self._request('PUT', url, data=putdata, params=kwargs)
        else:
            response = self._request('PUT', url, json=putdata, params=kwargs)
        if issubclass(model, BaseModel):
            return model(**response)
        else:
            return model(response)

    def patch(self, patchdata, model, *args, form=False, **kwargs):
        url = '/'.join(map(str, args))
        if form:
            response = self._request('PATCH', url, data=patchdata, params=kwargs)
        else:
            response = self._request('PATCH', url, json=patchdata, params=kwargs)
        if issubclass(model, BaseModel):
            return model(**response)
        else:
            return model(response)

    def delete(self, *args, **kwargs):
        url = '/'.join(map(str, args))
        return self._request('DELETE', url, params=kwargs)



import logging
from collections import defaultdict


class FlowInconsistency(Exception):
    pass


class OlcaParameterSet(object):
    def __init__(self, param_set_json):
        self._psj = param_set_json
        self._params = defaultdict(dict)
        for i, p in enumerate(self._psj.get('parameters', [])):
            if p['@type'] != 'ParameterRedef':
                logging.warning('non-ParameterRedef ignored [%d] (%s)' % (i, self.name))
                continue
            cx = p['context']['@id']
            pname = p['name']
            pval = p['value']
            self._params[cx][pname] = pval

    @property
    def name(self):
        return self._psj['name']

    @property
    def is_baseline(self):
        return bool(self._psj['isBaseline'])

    @property
    def contexts(self):
        for k in self._params.keys():
            yield k

    def param_values(self, context):
        return {k: v for k, v in self._params[context].items()}


class OlcaProductSystemInterpreter(object):
    """
    A convenience layer for interacting with an OpenLCA product system object
    """
    def __init__(self, ps_json, parameter_set=None):
        """

        :param ps_json: supply the json object corresponding to the named product system
        """
        self._psj = ps_json
        self._targets = defaultdict(dict)
        self._flows = defaultdict(dict)  # this is for errorchecking only
        for k in self._psj['processLinks']:
            parent = k['process']['@id']
            flow = k['flow']['@id']
            internal_id = k['exchange']['internalId']
            target = k['provider']['@id']
            self._targets[parent][internal_id] = target
            self._flows[parent][internal_id] = flow
        self._param_sets = [OlcaParameterSet(pset) for pset in self._psj.get('parameterSets', [])]
        if len(self._param_sets) > 1:
            if parameter_set is None:
                logging.warning('Using first parameter set %s' % self._param_sets[0].name)
                parameter_set = 0
            elif isinstance(parameter_set, int):
                if parameter_set >= len(self._param_sets):
                    logging.warning('Invalid parameter set specified (%d); using %s' % (parameter_set,
                                                                                        self._param_sets[0].name))
                    parameter_set = 0
            else:
                try:
                    parameter_set = next(i for i, p in enumerate(self._param_sets) if p.name == parameter_set)
                except StopIteration:
                    logging.warning('Invalid parameter set specified (%s); using %s' % (parameter_set,
                                                                                        self._param_sets[0].name))
                    parameter_set = 0
        else:
            parameter_set = 0
        self._param_set = parameter_set

    def _get_param_set(self, key):
        if key is None:
            return self._param_sets[self._param_set]
        elif isinstance(key, int):
            return self._param_sets[key]
        elif isinstance(key, str):
            try:
                return next(p for p in self._param_sets if p.name == key)
            except StopIteration:
                pass
        raise KeyError(key)

    @property
    def uuid(self):
        return self._psj['@id']

    @property
    def name(self):
        return self._psj['name']

    @property
    def processes(self):
        for p in self._psj['processes']:
            yield p['@id']

    @property
    def ref_process(self):
        return self._psj['refProcess']['@id']

    @property
    def ref_exchange(self):
        return self._psj['refExchange']['internalId']

    @property
    def parameter_sets(self):
        for s in self._param_sets:
            yield s.name

    def get_param_values(self, context, param_set=None):
        p = self._get_param_set(param_set)
        return p.param_values(context)

    def check_exchange_id(self, context, exchange_id):
        try:
            _ = self._targets[context][exchange_id]
            _ = self._flows[context][exchange_id]
            return True
        except KeyError:
            return False

    def get_target(self, context, exchange_id):
        return self._targets[context][exchange_id]

    def get_target_flow(self, context, exchange_id):
        """
        ps links specify anchor_flow, which must be used to build our model because bg engine expects it
        :param context:
        :param exchange_id:
        :return:
        """
        return self._flows[context][exchange_id]



from antelope.models import ResponseModel, EntityRef
from typing import Optional, List, Dict

UNRESOLVED_ANCHOR_TYPE = 'term'  # this is used when an anchor node's origin cannot be resolved


class Anchor(ResponseModel):
    """
    An anchor is either: a terminal node designation (i.e. origin + ref) or a context, and a descent marker.
    and cached LCIA scores

    Use FlowTermination.to_anchor(term, ..) to produce
    """
    node: Optional[EntityRef] = None
    anchor_flow: Optional[EntityRef] = None
    context: Optional[List[str]] = None
    descend: bool
    score_cache: Optional[Dict[str, float]] = None

    def __str__(self):
        """
        Replicate FlowTermination

        :return:
          '---:' = fragment I/O
          '-O  ' = foreground node
          '-*  ' = process
          '-#  ' - sub-fragment (aggregate)
          '-#::' - sub-fragment (descend)
          '-B ' - terminated background
          '--C ' - cut-off background
          '--? ' - ungrounded catalog ref

        :return:
        """
        if self.node:
            if self.node.entity_type == 'process':
                return '-*  '
            elif self.node.entity_type == UNRESOLVED_ANCHOR_TYPE:
                return '--? '
            else:
                if self.descend:
                    return '-#::'
                else:
                    return '-#  '
        elif self.context:
            if self.context == ['None']:
                return '-)  '
            else:
                return '-== '
        else:
            return '---:'

    @property
    def type(self):
        if self.node:
            return 'node'
        elif self.context:
            return 'context'
        else:
            return 'cutoff'

    @property
    def unit(self):
        if self.is_null:
            return '--'
        if self.node:
            if self.node.entity_type == 'fragment':
                return '####'
            return '****'
        return '::::'

    @property
    def is_null(self):
        return not bool(self.node or self.context)

    @classmethod
    def null(cls):
        return cls(descend=True)




from antelope import comp_dir

from .entities.fragments import Fragment
from .entities.anchors import UnCachedScore, UnresolvedAnchor
from .lcia_results import LciaResult


class FragmentInventoryDeprecated(Exception):
    """
    The "inventory" term for fragments is deprecated.  "unit_inventory" is now "unit_flows" and "inventory"
    is what it has always *meant*: "cutoffs".  And the old "cutoffs" is gone because it literally never worked.
    """
    pass


class CumulatingFlows(Exception):
    """
    when a fragment includes multiple instances of the reference flow having consistent (i.e. not complementary)
    directions. Not handled in subfragment traversal bc no valid test case
    """
    pass


class FragmentFlow(object):
    """
    A FragmentFlow is a an immutable record of a traversal query. essentially an enhanced NodeCache record which
    can be easily serialized to an antelope fragmentflow record.

    A fragment traversal generates an array of FragmentFlow objects.

    X    "fragmentID": 8, - added by antelope
    X    "fragmentStageID": 80,

    f    "fragmentFlowID": 167,
    f    "name": "UO Local Collection",
    f    "shortName": "Scenario",
    f    "flowID": 371,
    f    "direction": "Output",
    f    "parentFragmentFlowID": 168,
    f    "isBackground": false,

    w    "nodeWeight": 1.0,

    t    "nodeType": "Process",
    t    "processID": 62,

    *    "isConserved": true,
    *    "flowPropertyMagnitudes": [
      {
        "flowPropertyID": 23,
        "unit": "kg",
        "magnitude": 1.0
      }
    ]

    """
    '''
    @classmethod
    def from_antelope_v1(cls, j, query):
        """
        Need to:
         * create a termination
         * create a fragment ref
         * extract node weight
         * extract magnitude
         * extract is_conserved
        :param j: JSON-formatted fragmentflow, from a v1 .NET antelope instance.  Must be modified to include StageName
         instead of fragmentStageID
        :param query: an antelope v1 catalog query
        :return:
        """
        return cls(frag, magnitude, nw, term, conserved)

    @classmethod
    def ref_flow(cls, parent, use_ev):
        """

        :param parent:
        :param use_ev: required to create reference flows from fragment refs
        :return:
        """
        fragment = GhostFragment(parent, parent.flow, comp_dir(parent.direction))
        term = FlowTermination.null(fragment)
        return cls(fragment, use_ev, 1.0, term,
                   parent.is_conserved_parent)
    '''
    @classmethod
    def from_process_inventory(cls, query, process, ref_flow, elementary=False, exterior=False):
        """
        expands the named process's inventory into a list of FragmentFlows amenable to
        LCIA computation.

        generate FragmentFlows of the process's intermediate exchanges.  the ensuing list should be suitable to feed
        directly into frag_flow_lcia() to generate a contribution analysis
        :param query: only requires .get(), so an implementation will work fine
        :param process:
        :param ref_flow:
        :param elementary: [False] if True, also generate FFs for elementary exchanges (these will otherwise get
        computed via unobserved_exchanges()
        :param exterior: [False] if True, also generate FFs for non-elementary exterior exchanges
        :return:
        """
        rx = process.reference(ref_flow)
        parent = Fragment(process.external_ref, rx.flow, comp_dir(rx.direction),
                          exchange_value=1.0, observe=True)
        node = parent.observe_anchor(None, process, anchor_flow=rx.flow)
        ff = [cls(parent, 1.0, 1.0, node, False)]
        for ex in process.ad(ref_flow):  # returns exchangevalues
            n = query.get(ex.termination)
            cf = Fragment(n.external_ref, ex.flow, ex.direction, exchange_value=ex.value, observe=True,
                          parent=parent, anchor=n, anchor_flow=ex.flow)

            ff.append(cls(cf, ex.value, ex.value, cf.anch, False))
        if exterior:
            for ex in process.bf(ref_flow=ref_flow):
                if ex.termination.elementary and not elementary:
                    continue
                # still include non-elementary cutoffs

                cf = Fragment(ex.flow.uuid, ex.flow, ex.direction, exchange_value=ex.value, observe=True,
                              parent=parent, anchor=ex.termination)
                ff.append(cls(cf, ex.value, ex.value, cf.anch, False))

        return ff

    def __init__(self, fragment, magnitude, node_weight, term, is_conserved, match_ev=None, match_term=None,
                 flow_conversion=1.0):
        """

        :param fragment:
        :param magnitude:
        :param node_weight: flow (or balance) magnitude * flow conversion / anchor inflow magnitude
        :param term:
        :param is_conserved:
        :param match_ev:
        :param match_term:
        :param flow_conversion: [1.0] stored in the FragmentFlow for information purposes only. Negative value
         indicates a direction change at the anchor (driven anchor)
        """
        # TODO: figure out how to cache + propagate scenario applications through aggregation ops
        self.fragment = fragment
        self.magnitude = magnitude
        self.node_weight = node_weight
        self.flow_conversion = flow_conversion
        self.term = term
        self.is_conserved = is_conserved
        self._subfrags_params = ()
        self.match_scenarios = (match_ev, match_term)

    @property
    def subfragments(self):
        if self.term.is_subfrag:  # and (self.term.descend is False):
            try:
                return self._subfrags_params[0]
            except IndexError:
                return []
        return []

    @property
    def subfragment_scenarios(self):
        try:
            return self._subfrags_params[1]
        except IndexError:
            return None

    def aggregate_subfragments(self, subfrags, scenarios=None):
        """
        We need to save the full traversal specification (
        :param subfrags:
        :param scenarios:
        :return:
        """
        self._subfrags_params = (subfrags, scenarios)

    def scale(self, x):
        self.node_weight *= x
        self.magnitude *= x

    @property
    def name(self):
        return self.term.name

    def screen_name(self, length=80):
        """
        auto-compact for display

        :return:
        """
        name = self.name
        if len(name) > length:
            name = name[:(length - 18)] + '....' + name[-14:]
        return name

    def __str__(self):
        return '%.5s  %10.3g [%6s] %s %s' % (self.fragment.uuid, self.magnitude, self.fragment.direction,
                                             self.term, self.name)

    def __add__(self, other):
        if isinstance(other, FragmentFlow):
            if other.fragment.uuid != self.fragment.uuid:
                raise ValueError('Fragment flows do not belong to the same fragment')
            mag = other.magnitude
            nw = other.node_weight
            if not self.term == other.term:
                raise ValueError('These fragment flows are differently terminated')

            if mag * self.node_weight != (self.magnitude * nw):  # formally if m*N/M*n != 1.0:
                raise ValueError('These fragment flows cannot be combined because their implicit evs do not match')
            conserved = self.is_conserved and other.is_conserved

            mod_mag = mag * other.flow_conversion / self.flow_conversion
        else:
            raise TypeError("Don't know how to add type %s to FragmentFlow\n %s\n to %s" % (type(other), other, self))
        # don't check unit scores-- ?????
        new = FragmentFlow(self.fragment, self.magnitude + mod_mag, self.node_weight + nw,
                           self.term, conserved, flow_conversion=self.flow_conversion)
        return new

    def __eq__(self, other):
        """
        FragmentFlows are equal if they have the same fragment and termination.  Formerly magnitude too but why?
        answer why: because if not, then two traversals from two different scenarios can appear equal
        answer why not: because of LciaResult issues, presumably- let's put this on the list for later
        :param other:
        :return:
        """
        if not isinstance(other, FragmentFlow):
            return False
        return self.fragment == other.fragment and self.term == other.term  # and self.node_weight == other.node_weight

    def __hash__(self):
        return hash(self.fragment)

    @property
    def ref_unit(self):
        return self.fragment.flow.unit


def frag_flow_lcia(fragmentflows, quantity_ref, scenario=None, **kwargs):
    """
    Recursive function to compute LCIA of a traversal record contained in a set of Fragment Flows.
    Note: refresh is no longer supported during traversal
    :param fragmentflows:
    :param quantity_ref:
    :param scenario: necessary if any remote traversals are required
    :param ignore_uncached: [True] whether to allow zero scores for un-cached, un-computable fragments
    :return:
    """
    result = LciaResult(quantity_ref, scenario=str(scenario))
    _first_ff = True
    for ff in fragmentflows:
        _recursive_remote = False
        if ff.term.is_null:
            continue

        node_weight = ff.node_weight
        if node_weight == 0:
            continue

        if ff.term.direction == ff.fragment.direction:
            # if the directions collide (rather than complement), the term is getting run in reverse
            node_weight *= -1

        # if we have subfragments, use them
        if len(ff.subfragments) == 0:  # always true for: contexts, processes, remote subfragments
            try:
                v = ff.term.score_cache(quantity=quantity_ref, **kwargs)

                # if we reach here, then we have successfully retrieved a cached unit score and we are done
                if not v.is_null:
                    result.add_summary(ff.fragment.uuid, ff, node_weight, v)

                _first_ff = False
                continue

            except UnresolvedAnchor:
                result.add_missing(ff.fragment.uuid, ff.term.term_node, node_weight)
                _first_ff = False
                continue

            except UnCachedScore:
                # a subfragment with no stored subfragments and no cached score: we gotta ask
                v = ff.term.term_node.fragment_lcia(quantity_ref, scenario=scenario)
                _recursive_remote = True

        else:
            v = frag_flow_lcia(ff.subfragments, quantity_ref, scenario=ff.subfragment_scenarios, **kwargs)
            if v.is_null:
                continue

        # if we arrive here, we have a unit score from a subfragment

        if ff.term.descend:
            if v.has_summaries:
                for k in v.keys():
                    c = v[k]
                    result.add_summary(k, c.entity, c.node_weight * node_weight, c.internal_result)
            else:
                result.add_summary(ff.fragment.uuid, ff, node_weight, v)
        else:
            result.add_summary(ff.fragment.uuid, ff, node_weight, v)

        if _first_ff and _recursive_remote:
            return result  # bail out -- not sure why though- we need to be able to continue w/ child flows
        _first_ff = False
    return result



import json
import os
from collections import defaultdict

from antelope import UnknownOrigin

from .archives import InterfaceError, index_archive, update_archive, create_archive

# from .foreground import LcForeground
from .catalog_query import INTERFACE_TYPES, NoCatalog, zap_inventory

# from .providers import create_archive
from . import herd_factory


class ResourceInvalid(Exception):
    """
    resource points to an invalid filename
    """
    pass


class LcResource(object):
    """
    This is a record that links a semantic reference to a physical data source, and specifies the capabilities
    (and someday, access limitations) of the data source.

    The LcResource serializes to a json file with the following format:
    { ref: [ { "dataSource": source, "dataSourceType": ds_type, .... }, ... ] }
    where ref is the semantic reference.

    """
    @classmethod
    def from_archive(cls, archive, interfaces, source=None, **kwargs):
        source = source or archive.source
        ref = archive.ref
        ds_type = archive.__class__.__name__  # static flag indicates whether archive is complete
        kwargs.update(archive.init_args)
        static = kwargs.pop('static', archive.static)
        res = cls(ref, source, ds_type, interfaces=interfaces, static=static, preload_archive=archive, **kwargs)

        return res

    @classmethod
    def from_dict(cls, ref, d):
        """
        Returns a single LcResource loaded from a dict.  only required field is 'dataSourceType'.
        other fields are passed to the constructor and either interpreted directly or added as supplemental args

        If 'dataSource' is not present, one had better hope that url is present in the dict to download the source
        :param ref:
        :param d:
        :return:
        """
        source = d.pop('dataSource', None)
        ds_type = d.pop('dataSourceType')

        # patch to deal with changing Background extension handling-- this is WACKO
        filetype = d.pop('filetype', None)
        if filetype is not None:
            if not source.endswith(filetype):
                source += filetype

        return cls(ref, source, ds_type, **d)

    @classmethod
    def from_file(cls, file):
        """
        generates LcResources contained in the named file, sorted by increasing priority.  The filename and
        the reference must be the same.
        :param file:
        :return: an ordered list of resources
        """
        ref = os.path.basename(file)
        with open(file, 'r') as fp:
            j = json.load(fp)
        return cls.from_json(j, ref)

    @classmethod
    def from_json(cls, j, ref):
        return sorted([cls.from_dict(ref, d) for d in j[ref]], key=lambda x: x.priority)

    def _instantiate(self, catalog=None):
        """
        Instantiate the archive described by the current resource.  Several steps:
         - download file if applicable
         - rename relative path to absolute path
         - instantiate the archive
         - load any cached data into the archive
         - override static spec to be consistent with archive's own spec
         - load_all() if static
         - apply_config()

        :param catalog: required for: file downloads, foreground interfaces, local paths + caches
        :return:
        """
        if self.source is None:
            # download
            if catalog is None:
                raise NoCatalog('Remote resource encountered')
            if 'download' in self._args:
                print('Downloading from %s' % self._args['download']['url'])
                self._source = catalog.download_file(localize=True, **self._args['download'])
                self.write_to_file(catalog.resource_dir)  # update resource file
            else:
                raise AttributeError('Resource has no source specified and no download information')

        if self.source.startswith('$CAT_ROOT'):
            try:
                src = catalog.abs_path(self.source)
            except AttributeError:
                raise NoCatalog('Relative path encountered but no catalog supplied')
        else:
            src = self.source

        kwargs = {**self.init_args}

        if 'foreground' in self.interfaces:
            kwargs['catalog'] = catalog

        if self.ds_type.lower() == 'ecoinventlcia':
            # this is a GIANT HACK
            # we need to bring along a local ecoinvent archive to lookup flow reference qtys-
            # Partial solution to the hack: reference entity is QUERYABLE from a basic query which the catalog
            # can provide- still need to specify
            ei_ref = '.'.join(['local', 'ecoinvent', kwargs['version']])
            try:
                res = catalog.get_resource(ei_ref, iface='exchange', strict=False)
                res.check(catalog)
                if hasattr(res.archive, 'load_flows'):
                    res.archive.load_flows()
                kwargs['ei_archive'] = res.archive
            except UnknownOrigin:
                pass

        try:
            self._archive = create_archive(src, self.ds_type, factory=herd_factory,
                                           ref=self.origin, **kwargs)
        except FileNotFoundError as e:
            raise ResourceInvalid('%s: %s' % (self.origin, e.filename))

        if catalog is not None and catalog.check_cache(self.source):
            update_archive(self._archive, catalog.cache_file(self.source))
        self._static |= self._archive.static
        if self.static and self.ds_type.lower() != 'json':
            self._archive.load_all()  # static json archives are loaded on open- load_all() would be redundant

    @property
    def is_loaded(self):
        return self._archive is not None  # and self._archive._loaded ??

    def remove_archive(self):
        self._archive = None

    def check(self, catalog):
        if self._archive is None:
            # TODO: try/catch exceptions or return false
            print('QQQQQQQQQQQQQQQQQQ %s QQQQQQQQQQQQQQQQQQ' % self.origin)
            self._instantiate(catalog)
            self.apply_config(catalog)  # can't remember why I set this to happen recurrently- but it's no good
            ''' # on second thought, I dont think we want to do this-- catalog_names is a mapping of ref to source
            if catalog:
                for name in self.archive.catalog_names:
                    if name not in catalog.origins:
                        catalog.add_nickname(name, self.origin)
            '''

        return True

    def save(self, catalog):
        self.write_to_file(catalog.resource_dir)

    def make_index(self, index_file, force=True, save=True):
        self.check(None)
        self._archive.load_all()

        the_index = index_archive(self._archive, index_file, force=force, save=save)

        return the_index

    def make_cache(self, cache_file):
        # note: do not make descendant
        self._archive.write_to_file(cache_file, complete=True, gzip=True)
        print('Created archive of %s containing:' % self._archive)
        self._archive.check_counter()

    def make_interface(self, iface):
        return self._archive.make_interface(iface)

    def apply_config(self, catalog=None):
        # if len(self._config) == 0:  # NOW we don't even need to alter blackbook!!!
        #    return
        # print('Applying stored configuration')
        try:
            self._archive.make_interface('configure').apply_config(self._config)
        except InterfaceError:
            pass
        # we are moving this below apply_config to allow the archive to add/edit hints, which it can do because
        # it receives the authentic config dict as an argument
        if catalog is not None:
            if 'hints' in self._config:
                catalog.lcia_engine.apply_hints(self._archive.catalog_names, self._config['hints'])

    def remove_interface(self, iface):
        iface = zap_inventory(iface)  # don't warn when interpreting resource specifications
        if iface in self._interfaces:
            self._interfaces.remove(iface)

    def add_interface(self, iface):
        iface = zap_inventory(iface)  # don't warn when interpreting resource specifications
        if iface in INTERFACE_TYPES:
            self._interfaces.add(iface)

    def _normalize_interfaces(self, interfaces):
        """
        Ensures that:
         - interfaces spec can be string or list
         X 'basic' appears -- we don't want a basic interface on every resource
        :param interfaces:
        :return:
        """
        # self.add_interface('basic')
        if interfaces is None:
            return
        if isinstance(interfaces, str):
            self.add_interface(interfaces)
        else:
            for k in interfaces:
                self.add_interface(k)

    def __init__(self, origin, source, ds_type, interfaces=None, privacy=0, priority=50, static=False,
                 preload_archive=None, config=None, **kwargs):
        """

        :param origin: semantic reference to data origin
        :param source: physical data source; 'None' allowed if 'downloadLink' argument provided
        :param ds_type: data source type
        :param interfaces: list which can include 'entity', 'foreground', or 'background'. Default 'foreground'
        :param privacy: Ignored / No longer used.
        :param priority: [50] priority level.. numeric (nominally 0-100), lowest priority resource is loaded first
        :param static: [False] if True, load_all() after initializing
        :param preload_archive: [None] use to assign an existing archive
        :param config: ConfigureInterface specifications
        :param kwargs: additional keyword arguments to constructor. Some interesting ones:
          download: a dict containing 'url' and optional 'md5sum' fields
          prefix: often used when accessing zipped archives
          token: a jwt used for authenticating to xdb
          'options': popped with values used as kwargs (per ResourceSpec)


        """
        '''
        if not os.path.exists(source):
            raise EnvironmentError('%s not found' % source)
        '''

        self._archive = preload_archive

        options = kwargs.pop('options', dict())
        kwargs.update(options)

        self._org = origin
        if source is None:
            if 'download' not in kwargs:
                raise KeyError('Resource must be initialized with either source or download')
        self._source = source
        self._type = ds_type
        self._static = static

        self._issaved = False

        self._interfaces = set()
        self._normalize_interfaces(interfaces)

        if 'add_interfaces' in kwargs:
            self._normalize_interfaces(kwargs['add_interfaces'])

        self._priority = int(priority)

        self._internal = kwargs.pop('_internal', False)

        self._config = defaultdict(set)

        if config:
            for k, v in config.items():
                for q in v:
                    self._add_config(k, *q)

        self._args = kwargs

    def __repr__(self):
        flags = ['']
        if self.internal:
            flags.append('_int')
        if self.static:
            flags.append('static')
        if self._archive is not None:
            flags.append('loaded ')
        if len(self._config) > 0:
            flags.append('%d cfg' % len(self._config))
        fgs = ' '.join(flags)

        return 'LcResource(%s, dataSource=%s:%s, %s [%d]%s)' % (self.origin, self.source, self.ds_type,
                                                                [k for k in self.interfaces], self.priority, fgs)

    def exists(self, path):
        if path is None:
            return False
        filename = os.path.join(path, self.origin)
        if os.path.exists(filename):
            try:
                with open(filename, 'r') as fp:
                    j = json.load(fp)
            except json.JSONDecodeError:
                return False

            if any([self.matches(k) for k in j[self.origin]]):
                return True
        return False

    @property
    def archive(self):
        return self._archive

    @property
    def is_saved(self):
        return self._issaved

    @property
    def origin(self):
        return self._org

    @property
    def source(self):
        return self._source

    @property
    def ds_type(self):
        return self._type

    @property
    def interfaces(self):
        for k in self._interfaces:
            yield k

    @property
    def internal(self):
        return self._internal

    @property
    def priority(self):
        return self._priority

    @property
    def static(self):
        return self._static or self.ds_type.lower() == 'json'

    @property
    def init_args(self):
        return self._args

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, config_dict):
        if self.archive is not None:
            try:
                cf = self.archive.make_interface('configure')
                for cfg, cfgs in config_dict.items():
                    for args in cfgs:
                        cf.check_config(cfg, args)
            except InterfaceError:
                pass
        # apply config without replacing it blindly
        for k, v in config_dict.items():
            for t in v:
                self._config[k].add(tuple(t))

    def satisfies(self, ifaces):
        if ifaces is None:
            return True
        if isinstance(ifaces, str):
            ifaces = [ifaces]
        for i in ifaces:
            # if i == 'basic':
            #     return True
            if i in self._interfaces:
                return True
        return False

    def add_hint(self, hint_type, term, matches, catalog=None):
        if hint_type in ('context', 'flowable', 'quantity'):
            self._add_config('hints', hint_type, term, matches)
            if catalog:
                catalog.lcia_engine.apply_hints(self._archive.catalog_names, [(hint_type, term, matches)])
                self.save(catalog)
            else:
                print('Hint is neither saved nor applied')
        else:
            raise TypeError('Invalid hint type %s' % hint_type)

    def _add_config(self, config, *args):
        """
        does no validation
        :param config:
        :param args:
        :return:
        """
        self._config[config].add(args)

    def _serialize_config(self):
        j = dict()
        for k, v in self._config.items():
            j[k] = sorted([list(g) for g in v], key=lambda x: x[0])
        return j

    def export_config(self):
        return self._serialize_config()

    def serialize(self, stripped=False):
        if stripped:
            j = {"static": self.static,
                 "priority": self.priority}
        else:
            j = {
                "dataSource": self.source,
                "dataSourceType": self.ds_type,
                "interfaces": sorted([k for k in self.interfaces]),
                "priority": self.priority,
                "static": self.static
            }
            if self.internal:
                j['_internal'] = True
        j.update(self._args)
        j['config'] = self._serialize_config()
        return j

    def matches(self, k):
        """
        Pretty cheesy.  When we serialize a set of resources, we need to make sure not to include self twice.  To
        make the comparison concrete, use a serialized resource as input.

        We were using dataSource as a unique identifier for resource entries; but the introduction of download links
         breaks that because a downloadable resource has no source until it's been downloaded.
         The solution is to fallback to download.url ONLY IF the resource has no source specified.
        :param k: a serialized LcResource
        :return:
        """
        if k['dataSource'] is not None and self.source is not None:
            return k['dataSource'] == self.source
        return k['download']['url'] == self._args['download']['url']

    def write_to_file(self, path, assign_ref=None, apply_config=None):
        """
        Adds the resource to a file whose name is the resource's semantic reference. If the same datasource is
        already present in the file, replace it with the current resource.  otherwise append.
        :param path: directory to store the resource file.
        :param assign_ref: assign this ref instead of the resource's current ref
        :param apply_config: overwrites configuration with supplied dict
        :return:
        """
        if assign_ref is None:
            assign_ref = self.origin
        if apply_config is not None:
            self.config = apply_config  # tests configuration before storing it
        if not os.path.isdir(path):
            if os.path.exists(path):
                raise ValueError('Please provide a directory path')
            os.makedirs(path)


        filename = os.path.join(path, assign_ref)
        if os.path.exists(filename):
            try:
                with open(filename, 'r') as fp:
                    j = json.load(fp)
            except json.JSONDecodeError:
                j = {assign_ref: []}

            resources = [k for k in j[assign_ref] if not self.matches(k)]
            resources.append(self.serialize())
        else:
            resources = [self.serialize()]
        with open(os.path.join(path, assign_ref), 'w') as fp:
            json.dump({assign_ref: resources}, fp, indent=2)
        self._issaved = True



from synonym_dict import SynonymDict, SynonymSet, MergeError
from antelope import convert, ConversionError


class QuantityUnitMismatch(Exception):
    pass


class QuantityAlreadySet(Exception):
    pass


class ValidationError(Exception):
    pass


class QuantitySynonyms(SynonymSet):
    """
    QuantitySynonyms are string terms that all refer to the same quantity of measure. They must all have the same
    unit, because they are used to define the unit of measure of flowables.  To repeat: quantity instances that have
    the same dimensionality but different units (e.g. kWh and MJ) are NOT SYNONYMS but distinct quantities. The
    LciaEngine should be able to handle conversions between these kinds of quantities.
    """

    def _check_incoming_unit(self, quantity):
        unit = quantity.unit
        if unit is None:
            return
        if not isinstance(unit, str):
            raise AttributeError(quantity, 'unit-str')
        if self._unit is None:
            if unit is not None:
                self._unit = unit
            return
        if unit == self.unit:
            return
        if self._quantity is not None:
            try:
                cv = convert(self._quantity, from_unit=unit, to=self.unit)
                if cv == 1.0:
                    return
                print('Unit conversion mismatch %s->%s = %g' % (unit, self.unit, cv))
            except ConversionError:
                print('%s: Unit conversion KeyError on %s' % (self._quantity.name, unit))

        raise QuantityUnitMismatch('incoming %s (set %s)' % (unit, self.unit))

    @classmethod
    def new(cls, quantity):
        return cls(quantity.name, quantity=quantity)

    def _validate_quantity(self, quantity):
        if not hasattr(quantity, 'entity_type'):
            raise AttributeError(quantity)
        if quantity.entity_type != 'quantity':
            raise TypeError(quantity)
        if not hasattr(quantity, 'unit'):
            raise AttributeError(quantity, 'unit')
        self._check_incoming_unit(quantity)
        return True

    def __init__(self, *args, quantity=None, unit=None, **kwargs):
        self._quantity = None
        super(QuantitySynonyms, self).__init__(*args, **kwargs)
        self._unit = unit
        if quantity is not None:
            self.quantity = quantity

    @property
    def quantity(self):
        return self._quantity

    @quantity.setter
    def quantity(self, item):
        if self.quantity is None:
            if self._validate_quantity(item):
                self._quantity = item
                self._unit = item.unit
                for term in item.quantity_terms():
                    self.add_term(term)
            else:
                raise ValidationError('Quantity fails validation for uncaught reason (%s)' % item)
        else:
            raise QuantityAlreadySet

    def _save_synonyms(self, other):
        """
        adds to other's synonym set-
        :param other:
        :return:
        """
        for k in other.terms:
            self._quantity.add_synonym(k)

    def add_child(self, other, force=False):
        if not isinstance(other, QuantitySynonyms):
            raise TypeError('Child set is not a Quantity synonym set (%s)' % type(other))
        self._check_incoming_unit(other)
        if self._quantity is None:
            self.quantity = other.quantity
        elif self._quantity.is_entity:
            self._save_synonyms(other)
        super(QuantitySynonyms, self).add_child(other, force=force)

    @property
    def object(self):
        return self._quantity

    @property
    def unit(self):
        return self._unit

    def serialize(self):
        d = super(QuantitySynonyms, self).serialize()
        if self.unit is not None:
            d['unit'] = self.unit

        return d


class QuantityManager(SynonymDict):

    _entry_group = 'Quantities'
    _syn_type = QuantitySynonyms

    _ignore_case = True

    def _add_from_dict(self, j):
        name = j['name']
        syns = j.pop('synonyms', [])
        unit = j.pop('unit', None)
        self.new_entry(name, *syns, unit=unit, merge=False)

    def find_matching_quantity(self, quantity):
        if isinstance(quantity, str):
            return self[quantity]
        for term in quantity.quantity_terms():
            if term in self:
                return self[term]
        raise KeyError(quantity)

    def add_quantity(self, quantity):
        """
        Prunes terms on quantity unit mismatch
        :param quantity:
        :return:
        """
        new_q = QuantitySynonyms.new(quantity)
        try:
            self.add_or_update_entry(new_q, merge=True, create_child=True)
        except QuantityUnitMismatch:
            # want to print a warning
            # if we get QuantityUnitMismatch, then that means we didn't get a MergeError-i.e. there is only one match
            me = self.match_entry(*new_q.terms)
            print('!! Warning: new quantity %s [%s] has unit conflict with existing quantity %s [%s] ' % (new_q,
                                                                                                          new_q.unit,
                                                                                                          me, me.unit))
            pe = self.add_or_update_entry(new_q, merge=False, prune=True)
            print('   Added pruned quantity %s [%s]' % (pe.name, pe.unit))
        except MergeError:
            """
            we have (at least) two different quantities.  If the link is already known, we're done
            """
            new_terms = [c for c in new_q.terms if self._check_term(c) is None]
            for best in (quantity.external_ref, quantity.link, quantity.name):
                k = self._check_term(best)
                if k is not None:
                    try:
                        k.add_child(new_q)
                    except QuantityUnitMismatch:
                        continue
                    # we found one
                    for nt in new_terms:
                        self.add_synonym(best, nt)
                    return
            # we did not find one
            print('!! %s: Abandoning orphan terms %s' % (quantity, new_terms))

    def add_synonym(self, term, syn):
        super(QuantityManager, self).add_synonym(term, syn)
        ent = self._d[syn]
        # canonical quantity is always a ref
        for child in ent.children:
            if child.quantity is not None and child.quantity.is_entity:
                child.quantity.add_synonym(str(syn).strip())

    def __getitem__(self, item):
        if hasattr(item, 'link'):
            item = item.link
        return super(QuantityManager, self).__getitem__(item)



from antelope import BaseEntity, CatalogRef, EntityNotFound
from antelope.models import Entity, FlowEntity, ReferenceExchange

import re


class XdbReferenceRequired(Exception):
    """
    Straight-up entities have no capabilities
    """
    pass


class XdbEntity(BaseEntity):

    is_entity = True  # haha, all lies!

    def __init__(self, model):
        """
        XdbEntity is an ephemeral object, basically a way of storing pydantic models PRIOR to their getting made into
        refs (which is done by this class's make_ref() process)

        XdbEntities are instantiated by the client on any occasion the client receives info on entities back from
        the remote server: in XdbClient._fetch(), and in processes() flows() quantities() and anything to do with
        exchanges.

        The return objects are immediately used as arguments for BasicQuery.make_ref() or CatalogQuery.make_ref(),
        either of which calls this class's make_ref() with the query as argument.  Then the make_ref() is responsible
        for constructing the fully-featured reference object that is stored in the local archive.

        Must supply the pydantic model that comes out of the query, and also the archive that stores the ref
        :param model:
        """
        assert issubclass(type(model), Entity), 'model is not a Pydantic Entity (%s)' % type(model)
        self._model = model
        self._ref = None

    @property
    def ref(self):
        if self._ref is None:
            raise XdbReferenceRequired
        return self._ref

    @property
    def reference_entity(self):
        raise XdbReferenceRequired

    @property
    def entity_type(self):
        return self._model.entity_type

    @property
    def origin(self):
        return self._model.origin

    @property
    def external_ref(self):
        return self._model.entity_id

    def properties(self):
        for k in self._model.properties:
            yield k

    def __setitem__(self, key, value):
        if self._ref:
            self._ref[key] = value
        self._model.properties[key] = value

    def __getitem__(self, item):
        if self._ref:
            return self._ref[item]
        return self._model.properties[item]

    def make_ref(self, query):
        if self._ref is not None:
            return self._ref

        args = {k: v for k, v in self._model.properties.items()}
        if self.entity_type == 'quantity' and 'unit' in args:
            args['reference_entity'] = args.pop('unit')
        elif self.entity_type == 'flow':
            if 'referenceQuantity' in args:
                rq = args.pop('referenceQuantity')
                try:
                    args['reference_entity'] = query.get_canonical(rq)
                except EntityNotFound:
                    args['reference_entity'] = query.get(rq)
            if isinstance(self._model, FlowEntity):
                args['context'] = self._model.context
                args['locale'] = self._model.locale
                # query._tm.add_context(self._model.context, origin=self._model.origin)  # this is somewhat cursed
        elif self.entity_type == 'process':
            if 'referenceExchange' in args:
                # we cannot synthesize RxRefs prior to the existence of the ProcessRef. sorry.
                rxs = args.pop('referenceExchange')
                try:
                    args['referenceExchange'] = [ReferenceExchange(**k) for k in rxs]
                except TypeError:
                    print(self.link)
                    print(rxs)
                    raise

        if self.origin != query.origin:
            args['masquerade'] = self.origin

        ref = CatalogRef.from_query(self.external_ref, query, self.entity_type, **args)
        if ref.entity_type == 'flow':
            if any(bool(re.search('carbon.dioxide', k, flags=re.I)) for k in ref.synonyms):
                print('%s ***** CO2' % ref.link)
                ref.is_co2 = True

        self._ref = ref
        return ref

    def has_lcia_engine(self):
        if self._ref is not None:
            return self._ref.has_lcia_engine()
        return False



"""
ChatGPT did not unfortunately succeed
"""

import ast
import hashlib


def normalize_expression(node):
    """ Normalize expressions to make some commutative operations consistent. """
    if isinstance(node, ast.BinOp) and isinstance(node.op, (ast.Add, ast.Mult)):
        left_hash = hash_ast(node.left)
        right_hash = hash_ast(node.right)
        if left_hash > right_hash:
            node.left, node.right = node.right, node.left  # Swap to maintain order
    return node


def hash_ast(node):
    """ Hash an AST node into a digest that uniquely represents its structure and contents. """
    hasher = hashlib.sha256()

    def hash_node(n):
        if isinstance(n, ast.AST):
            hasher.update(type(n).__name__.encode('utf-8'))
            for field, value in ast.iter_fields(n):
                if field == 'ctx':
                    continue  # Ignore context field
                hasher.update(field.encode('utf-8'))
                hash_node(value)
        elif isinstance(n, list):
            for item in n:
                hash_node(item)
        else:
            hasher.update(str(n).encode('utf-8'))
    hash_node(node)
    return hasher.hexdigest()


def find_common_subexpressions(formulas):
    subexpressions = {}
    for i, formula in enumerate(formulas):
        tree = ast.parse(formula, mode='eval')
        normalize_expression(tree.body)  # Normalize the entire expression tree
        for node in ast.walk(tree):
            if not isinstance(node, ast.Name):  # Focus on composite nodes
                node_hash = hash_ast(node)
                if node_hash in subexpressions:
                    subexpressions[node_hash].append((i, ast.unparse(node)))
                else:
                    subexpressions[node_hash] = [(i, ast.unparse(node))]
    return subexpressions


"""
# Find common subexpressions
common_subexpressions = find_common_subexpressions(formulas)
for h, occurrences in common_subexpressions.items():
    if len(occurrences) > 1:
        print(f"Common subexpression found in: {occurrences}")
"""



"""
FileAccessor, for standardizing access to antelope resources on a filesystem having the following structure:

{DATA ROOT}/[origin]/[interface]/[ds_type]/[source_file]  - source
{DATA ROOT}/[origin]/[interface]/[ds_type]/config.json    - configuration

A filesystem having this structure will enable automatic registration of resources, taking origin, interface, and
ds_type from the directory structure and the source+config files as-discovered by traversing the filesystem.

Should probably alter gen_sources to only return the one "best" source per path, instead of generating all of them.
[One] problem with that is that, for instance, for Tarjan background I would need to ignore files that end with
ORDERING_SUFFIX, but I can't import anything from antelope_background without creating a dependency loop.

For now, generating the sources is probably fine.

"""

import os
import json
from .lc_resource import LcResource
from antelope import BackgroundRequired
from shutil import copy2


DEFAULT_PRIORITIES = {
    'exchange': 20,
    'quantity': 20,
    'index': 50,
    'background': 80
}


class FileAccessor(object):

    def __init__(self, load_path, prefix=None):
        self._path = os.path.abspath(load_path)  # this has the benefits of collapsing '..' and trimming trailing '/'
        self._origins = os.listdir(self._path)
        self._prefix = prefix

    @property
    def path(self):
        return self._path

    @property
    def origins(self):
        for k in self._origins:
            yield k

    def refresh(self):
        self._origins = os.listdir(self._path)

    @staticmethod
    def read_config(source):
        cfg = os.path.join(os.path.dirname(source), 'config.json')
        if os.path.exists(cfg):
            with open(cfg) as fp:
                config = json.load(fp)
        else:
            config = dict()
        return config

    @staticmethod
    def write_config(source, **kwargs):
        """
        Note: use the config argument add_interfaces=[list] to allow a resource to implement additional interfaces.
        :param source:
        :param kwargs:
        :return:
        """
        cfg = os.path.join(os.path.dirname(source), 'config.json')
        with open(cfg, 'w') as fp:
            json.dump(kwargs, fp, indent=2)

    def update_config(self, source, **updates):
        existing_cfg = self.read_config(source)
        existing_cfg.update(updates)
        self.write_config(source, **existing_cfg)

    def clear_configs(self, origin):
        opath = os.path.join(self.path, origin)
        for iface in os.listdir(opath):
            ipath = os.path.join(opath, iface)
            for ds_type in os.listdir(ipath):
                dc = os.path.join(ipath, ds_type, 'config.json')  # don't understand why this is an error
                if os.path.exists(dc):
                    os.remove(dc)

    def source(self, org, iface):
        """
        Return the "best" (right now: first) single source
        :param org:
        :param iface:
        :return:
        """
        return next(self.gen_sources(org, iface))

    def gen_sources(self, org, iface):
        iface_path = os.path.join(self._path, org, iface)
        if not os.path.exists(iface_path):
            return
        for ds_type in os.listdir(iface_path):
            ds_path = os.path.join(iface_path, ds_type)
            if not os.path.isdir(ds_path):
                continue
            for fn in os.listdir(ds_path):
                if fn == 'config.json':
                    continue
                # if we want to order sources, this is the place to do it
                source = os.path.join(ds_path, fn)
                yield source

    def create_resource(self, source, basic=False, **kwargs):
        if not source.startswith(self._path):
            raise ValueError('Path not contained within our filespace')
        rel_source = source[len(self._path)+1:]
        org, iface, ds_type, fn = rel_source.split(os.path.sep)  # note os.pathsep is totally different

        cfg = self.read_config(source)
        cfg.update(kwargs)
        priority = cfg.pop('priority', DEFAULT_PRIORITIES[iface])

        if self._prefix is not None:
            org = '.'.join([str(self._prefix), org])

        # do this last
        ifaces = set()
        ifaces.add(iface)
        if basic:
            ifaces.add('basic')
        # if iface == 'index':
        #     # TODO: WORKAROUND: automatically add 'basic' to index interfaces until we have it sorted out
        #     ifaces.add('basic')

        return LcResource(org, source, ds_type, interfaces=tuple(ifaces), priority=priority, **cfg)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return


class ResourceLoader(FileAccessor):
    """
    This class crawls a directory structure and adds all the specified resources to the catalog provided as
    an input argument.

    After loading all pre-initialized resources, the class runs check_bg on each origin to auto-generate a
    background ordering within the catalog's filespace if one does not already exist.
    """

    def _load_origin(self, cat, org, check, replace=True):
        """
        Crawls the data path at the specified origin and creates a resource for each source found.
        TODO: figure out whether basic should be a standalone resource // how to handle documentation
        :param org:
        :param check: Whether to instantiate the interfaces
        :return:
        """
        check_bg = False
        for iface in ('exchange', 'index', 'background', 'quantity'):
            for i, source in enumerate(self.gen_sources(org, iface)):
                res = self.create_resource(source)
                cat.add_resource(res, replace=replace)
                if check:
                    res.check(cat)
                    if iface == 'background':
                        check_bg = True
        if self._prefix:
            dest_org = '.'.join([self._prefix, org])
        else:
            dest_org = org
        try:
            next(cat.gen_interfaces(dest_org, 'basic'))
            valid = True
        except StopIteration:
            print('Warning: %s: no basic interface (add manually and save)' % dest_org)
            valid = False
        if check_bg:
            try:
                bg = cat.query(dest_org).check_bg()
            except BackgroundRequired:
                bg = False
            return valid and bg
        else:
            return valid

    def load_resources(self, cat, origin=None, check=False, replace=True):
        """
        For named origin (or all origins), load resources into the catalog. this will not open the resources, unless
        check=True
        :param cat:
        :param origin:
        :param check: [False] whether to load the resources and check background
        :param replace: [True] whether to replace an existing resource with the new one
        :return:
        """
        if origin is None:
            status = []
            for org in self.origins:
                status.append(self._load_origin(cat, org, check, replace=replace))
        else:
            status = self._load_origin(cat, origin, check, replace=replace)
        return status


class ResourceWriter(object):

    def __init__(self, resource, root_dir, origin=None, interface=None, **kwargs):
        self.res = resource
        self.fa = FileAccessor(root_dir)
        if origin is None:
            self._origin = self.res.origin
        else:
            self._origin = origin
        if interface is None:
            for k in ('exchange', 'quantity', 'index', 'background'):
                if k in self.res.interfaces:
                    interface = k
                    break
            if interface is None:
                raise AttributeError('no interfaces!')

        self.interface = interface
        self.add_interfaces = [k for k in self.res.interfaces if k != interface]
        self.write(**kwargs)

    @property
    def origin(self):
        return self._origin

    @property
    def target_path(self):
        return os.path.join(self.fa.path, self.origin, self.interface, self.res.ds_type)

    @property
    def target(self):
        return os.path.join(self.target_path, os.path.basename(self.res.source))

    def write(self, force=False, **kwargs):
        if os.path.exists(self.target):
            if not force:
                raise FileExistsError(self.target)
        os.makedirs(self.target_path, exist_ok=True)
        if self.origin not in self.fa.origins:
            self.fa.refresh()
        if os.path.exists(self.res.source):
            copy2(self.res.source, self.target)
        else:
            self.res.archive.write_to_file(self.target, domesticate=True)
        ex_cfg = self.res.serialize(stripped=True)
        ex_cfg.update(kwargs)
        self.fa.write_config(self.target, add_interfaces=self.add_interfaces, **ex_cfg)
        print('written to %s' % self.target)



"""
What do we want to test here?
 flow vs flow ref, cf vs quantity_relation, .. same vs different origins?

"""
import unittest

from .. import Qdb
from ...implementations.quantity import ConversionReferenceMismatch

origin = 'test.origin'


class TestQuantityRelation(unittest.TestCase):

    f4_mj_kg = 35

    @classmethod
    def setUpClass(cls) -> None:
        qdb = Qdb.new()
        qi = qdb.make_interface('quantity')

        f4 = qdb.new_flow('Another mass flow', 'mass')
        qi.characterize(f4.link, f4.reference_entity, 'net calorific value', value=cls.f4_mj_kg)

        cls.f4 = f4
        cls.f6 = qdb.new_flow('An energetic conservation flow', 'net calorific value')
        cls.qdb = qdb

    @property
    def f4ref(self):
        return self.f4.make_ref(self.qdb.query)

    @property
    def f6ref(self):
        return self.f6.make_ref(self.qdb.query)

    def test_profile(self):
        self.assertEqual(len(list(self.f4ref.profile())), 1)

    def test_fwd_cf(self):
        self.assertEqual(self.f4.cf(self.f6.reference_entity), self.f4_mj_kg)

    def test_fwd_q_cf(self):
        mass = self.f4.reference_entity
        self.assertEqual(mass.cf(self.f6), 0.0)
        self.assertEqual(mass.cf(self.f4), 1.0)

        eng = self.f6.reference_entity
        self.assertEqual(eng.cf(self.f4), 35.0)
        self.assertEqual(eng.cf(self.f6), 1.0)

    def test_fwd_qr(self):
        mass = self.f4.reference_entity
        self.assertEqual(mass.quantity_relation(self.f4.name, self.f4.reference_entity, None).value, 1.0)
        with self.assertRaises(ConversionReferenceMismatch):
            mass.quantity_relation(self.f6.name, self.f6.reference_entity, None)

        eng = self.f6.reference_entity
        self.assertEqual(eng.quantity_relation(self.f4.name, mass, None).value, self.f4_mj_kg)
        self.assertEqual(eng.quantity_relation(self.f6.name, self.f6.reference_entity, None).value, 1.0)

    def test_rev_cf(self):
        mass = self.f4.reference_entity
        eng = self.f6.reference_entity
        self.assertEqual(mass.quantity_relation(self.f4.name, eng, None).value,
                         1.0 / self.f4_mj_kg)
        with self.assertRaises(ConversionReferenceMismatch):
            eng.quantity_relation(self.f6.name, mass, None)





import unittest


from ..local import make_config
from ... import LcCatalog
from antelope import ExchangeRef
from antelope.refs.process_ref import ProcessRef
from antelope.refs.flow_ref import FlowRef

tc = make_config('traci')
traci_origin = next(tc.origins)


class DummyQuery:
    origin = 'test.dummy'
    def validate(self):
        return True


def generate_gwp_inventory(_cat):
    dq = DummyQuery()
    node = ProcessRef('aabbccdd', dq, Name='dummy GWP process')
    mass = _cat.get_canonical('mass')
    to_air = _cat.lcia_engine['to air']
    yield ExchangeRef(node, FlowRef(1, dq, reference_entity=mass, name='Carbon Dioxide'), 'Output', 2.2,
                      termination=to_air)
    yield ExchangeRef(node, FlowRef(2, dq, reference_entity=mass, name='Carbon Dioxide (biogenic)'), 'Output', 1.1,
                      termination=to_air)
    yield ExchangeRef(node, FlowRef(2, dq, reference_entity=mass, name='methane'), 'Output', 0.5,
                      termination=to_air)


class TestTraci(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.cat = LcCatalog.make_tester()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.cat.__del__()

    def test_0_register_traci(self):
        tc.register_all_resources(self.cat)
        self.assertIn(traci_origin, self.cat.origins)

    def test_1_load_traci(self):
        qs = list(self.cat.query(traci_origin).lcia_methods())
        self.assertEqual(len(qs), 10)

    def test_2_traci_factors(self):
        gwp = self.cat.query(traci_origin).get('Global Warming Air')
        cfs = list(gwp.factors())
        self.assertEqual(len(cfs), 91)

    def test_3_traci_lcia(self):
        gwp = self.cat.query(traci_origin).get('Global Warming Air')
        inv = list(generate_gwp_inventory(self.cat))
        self.assertEqual(gwp.do_lcia(inv).total(), 15.8)
        gwp['quell_biogenic_co2'] = True
        self.assertEqual(gwp.do_lcia(inv).total(), 14.7)


if __name__ == '__main__':
    unittest.main()



import logging
from collections import defaultdict
import os

from ..archives.term_manager import TermManager, NoFQEntry, DuplicateCharacterization, FactorConflict
from ..contexts import Context, NullContext
from .clookup import CLookup, SCLookup, FactorCollision

from antelope.flows.flow import flowname_is_biogenic

from synonym_dict import TermExists, FlowablesDict
# from synonym_dict.example_compartments.compartment_manager import InconsistentLineage  # this is not needed


'''
Switchable biogenic CO2:

* Biogenic CO2 is CO2, so the flowable used to store CFs is 124-38-9 always
* Because flowable is looked up within the quantity implementation, we can assign a synonym to the flow itself, and 
  watch for it
* Then the switch determines whether or not to quell the CF returned to the user, without changing the database
'''


DEFAULT_CONTEXTS = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data', 'contexts.json'))
NUM_DEFAULT_CONTEXTS = 37  # added fossil

DEFAULT_FLOWABLES = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data', 'flowables.json'))


class QuantityMasqueradeError(Exception):
    pass


class LciaEngine(TermManager):
    """
    This adds several abilities:
     * track flowables by origin
       - identify flowables thhat are not recognized
       - return flowables by origin (all, or only new)
     * lookup CFs based on context hierarchy
       - dist = 0: only exact matchh
       - dist = 1: match or subcompartments
       - dist = 2: match, subcompartments, or parent
       - dist = 3: .. or any parent up to NullContext
     * quell biogenic CO2 in quantity relation lookups
    """
    is_lcia_engine = True

    def _configure_flowables(self, flowables):
        """
        Setup local flowables database with flows that require special handling. Also loads the flowables file.

        When overriding this function, place the super() call between pre-load and post-load activities.
        :return:
        """
        if flowables is None:
            flowables = DEFAULT_FLOWABLES

        # FlowablesDict-- mainly to upsample CAS numbers for matching
        self._fm = FlowablesDict()

        self._fm.new_entry('carbon dioxide', '124-38-9')
        self._water = self._fm.new_entry('Water', '7732-18-5')

        # now load + merge from file
        self._fm.load(flowables)

        self._fq_map = {fb: set() for fb in self._fm.objects}

    def __init__(self, contexts=None, flowables=None, quantities=None,
                 quell_biogenic_co2=None,
                 strict_clookup=True,
                 **kwargs):
        """

        :param quantities:
        :param quell_biogenic_co2: DEPRECATED
        :param contexts:
        :param flowables:
        :param strict_clookup: [True] whether to prohibit multiple CFs for each quantity / flowable / context tuple
        :param kwargs: from TermManager: quiet, merge_strategy
        """
        if contexts is None:
            contexts = DEFAULT_CONTEXTS
        super(LciaEngine, self).__init__(contexts=contexts, flowables=None, quantities=quantities, **kwargs)

        # override flowables manager
        self._configure_flowables(flowables)

        # the CF lookup: allow hierarchical traversal over compartments [or, um, use a graph db..]
        self._cl_typ = {True: SCLookup,
                        False: CLookup}[strict_clookup]  #
        # another reverse mapping
        self._origins = set()
        self._fb_by_origin = defaultdict(set)  # maps origin to flowables having that origin
        self._fb_by_origin[None] = set(str(k) for k in self._fm.objects)

        if quell_biogenic_co2 is not None:
            print('''WARNING: quell_biogenic_co2 LciaEngine argument is DEPRECATED
It is now a PROPERTY SETTING on a quantity.''')

        self._factors_for_later = defaultdict(bool)

    def get_canonical(self, quantity):
        """
        We override this because here we are using canonical quantities and 'kg' is not a canonical quantity, but it
        always means mass
        :param quantity:
        :return:
        """
        if quantity == 'kg':  # for THIS UNIT ONLY
            quantity = 'mass'
        return super(LciaEngine, self).get_canonical(quantity)

    def save_for_later(self, quantity):
        qc = self.get_canonical(quantity)
        if qc is quantity:
            raise QuantityMasqueradeError(quantity)
        self._factors_for_later[qc] = quantity

    def __getitem__(self, item):
        """
        LciaEngine.__getitem__ retrieves a canonical context by more intensively searching for matches from a given
        context.  Adds foreign context's full name as synonym if one is affirmatively found.  If one is not found,
        returns the NullContext.

        None is returned as None, to represent 'unspecified' (i.e. accept all) as opposed to 'no context' which isa
        context (accept only matching). (as tested)

        :param item:
        :return: a matching context or NullContext.
        """
        if item is None:
            return None
        try:
            return self._cm.__getitem__(item)
        except KeyError:
            if isinstance(item, Context):
                return self._cm.find_matching_context(item)
            elif isinstance(item, tuple) and len(item) > 1:
                return self.__getitem__(item[1:])
            return NullContext

    def apply_hints(self, names, hints):
        """
        Hints should be
        :param names: iterable of catalog names to which context hints will apply
        :param hints: an iterable of term, canonical pairs, where term is the context as known in the origin, and
        canonical is the corresponding canonical context.
        :return:
        """
        orgs = list(names)
        for hint_type, term, canonical in hints:
            if term in self.synonyms(canonical):
                continue
            if hint_type == 'context':
                for org in orgs:
                    print('Applying context hint %s:%s => %s' % (org, term, canonical))
                    self._cm.add_context_hint(org, term, canonical)
            elif hint_type == 'quantity':
                print('Applying quantity hint %s -> %s' % (term, canonical))
                try:
                    self._qm.add_synonym(canonical, term)
                except TermExists:
                    c = self._qm[canonical]
                    t = self._qm[term]
                    if c is not t:
                        print('!X!X!X!X!X!X!X! conflict when applying hint %s->%s' % (term, canonical))
                        print('                canonical: %s = ' % c.link)
                        print('                collision: %s = ' % t.link)
                    # assert self._qm[canonical] is self._qm[term]
            elif hint_type == 'flowable':
                print('Applying flowable hint %s -> %s' % (term, canonical))
                try:
                    self.add_terms('flowable', canonical, term)
                except TermExists:
                    assert self._fm[canonical] == self._fm[term]
            else:
                raise ValueError('Unknown hint type %s' % hint_type)

    def add_synonym(self, existing_term, synonym):
        """

        :param existing_term: flowable or quantity (cx is TBD)
        :param synonym: string term to add
        :return:
        """
        synonym = str(synonym).strip()
        try:
            ent = self._fm[existing_term]
            self._add_to_existing_flowable(ent, (synonym,))
        except KeyError:
            try:
                ent = self._qm[existing_term]
                self._qm.add_synonym(ent, synonym)
            except KeyError:
                raise KeyError('No entry found for %s' % existing_term)

    @staticmethod
    def _flow_terms(flow):
        """
        This function was created because we don't want TermManagers getting confused about things like misassigned
        CAS numbers, because they don't have the capacity to store multiple CFs.  So we will save the full synonym
        list for the LciaEngine.
        :param flow:
        :return:
        """
        return flow.synonyms

    def _check_fb_map(self, fb_map):
        """
        flowable match curation:
         - We don't want to merge any incoming terms with "water" because it makes flow-based water footprinting
            impossible
        :param fb_map: the dict being used to manage
        :return:
        """
        if len(fb_map) > 1:
            fb_map.pop(self._water, None)

    def _add_to_existing_flowable(self, fb, new_terms):
        """
        Harvest biogenic co2 synonyms (not worth correcting for other biogenic substances??)
        :param fb:
        :param new_terms:
        :return:
        """
        for term in new_terms:
            self._fm.add_synonym(fb, term)

    def add_flow_terms(self, flow, merge_strategy=None):
        """
        Subclass handles two problems: tracking flowables by origin and biogenic CO2.

        Should probably test this shit

        Under our scheme, it is a requirement that the flowables list used to initialize the LciaEngine is curated.

        biogenic: if ANY of the flow's terms match the biogenic regex AND the flow is CO2, set its name
        :param flow:
        :param merge_strategy:
        :return:
        """
        fb = super(LciaEngine, self).add_flow_terms(flow, merge_strategy=merge_strategy)
        self._fb_by_origin[flow.origin].add(str(fb))
        if '124-38-9' in fb:
            flow.is_co2 = True
            try:
                bio = next(t for t in flow.synonyms if flowname_is_biogenic(t))
            except StopIteration:
                # no biogenic terms
                return fb
            flow.name = bio  # ensure that flow's name shows up with that term
            self._fb_by_origin[flow.origin].add(bio)
        return fb

    def add_quantity(self, quantity):
        """
        Here we are not picky about having duplicate quantities
        :param quantity:
        :return:
        """
        if quantity.entity_type != 'quantity':
            raise TypeError('Must be quantity type')
        if quantity.link not in self._qm:
            self._qm.add_quantity(quantity)
            assert quantity.link in self._qm
        return self._canonical_q(quantity)

    '''
    def get_canonical(self, quantity):
        """
        Override TermManager default to automatically add unknown quantities
        :param quantity:
        :return:
        """
        try:
            return self._canonical_q(quantity)
        except KeyError:
            if hasattr(quantity, 'entity_type') and quantity.entity_type == 'quantity' and not quantity.is_entity:
                print('Missing canonical quantity-- adding to LciaDb')
                self._catalog.register_quantity_ref(quantity)
                q_can = self._tm.get_canonical(quantity)
    '''

    def merge_quantities(self, first, second):
        """
        Absorb second into child of first.  Currently does not support remapping entries, so import_cfs(second)
        needs to be run again. Old factors will be left in so _fq_map still works.
        :param first:
        :param second:
        :return:
        """
        dom = self.get_canonical(first)
        add = self.get_canonical(second)
        self._qm.merge(dom, add)
        self.import_cfs(second)

    def _add_compartments(self, comps):
        """
        For LciaEngines, we use the default behavior of ContextManager which is to first try 'attach', then fallback
        to 'rename'
        :param comps:
        :return:
        """
        return self._cm.add_compartments(comps)

    '''  # This doesn't do anything useful
    def add_subcontext(self, context, prefix, sub, origin=None):
        """

        :param context: a canonical context
        :param prefix:
        :param sub: a new subcompartment
        :param origin: [None]
        :return:
        """
        if context is NullContext or context is None or context == ():
            return context
        cx_sub = tuple(context.as_list() + ['%s-%s' % (prefix, sub)])

        try:
            cx = self._cm[cx_sub]
        except (KeyError, InconsistentLineage):
            cx = self.add_context(cx_sub, origin=origin)
        return cx
    '''

    def import_cfs(self, quantity):
        """
        Given a quantity, import its CFs into the local database.  Unfortunately this is still going to be slow because
        every part of the CF still needs to be canonicalized. The only thing that's saved is creating a new
        Characterization instance.
        :param quantity: this is a TRUE quantity whose query points to the authentic origin.
        :return:
        """
        try:
            qq = self._canonical_q(quantity)
        except KeyError:
            qq = self.add_quantity(quantity)

        count = 0
        for cf in quantity.factors():
            count += 1
            # print(cf)
            try:
                fb = self._fm[cf.flowable]
            except KeyError:
                if quantity.has_lcia_engine():  # this is the native quantity, remember
                    fb = self._create_flowable(*quantity.query_synonyms(cf.flowable))
                else:
                    fb = self._create_flowable(cf.flowable)

            self.add_quantity(cf.ref_quantity)  # this may lead to the creation of non-converting quantities if units mismatch

            cx = self[cf.context]

            try:
                self._qassign(qq, fb, cf, context=cx)
            except FactorCollision:
                rq = self._canonical_q(cf.ref_quantity)
                ex_cf = self._find_exact_cf(qq, fb, cx, quantity.origin)
                if rq != ex_cf.ref_quantity:
                    for loc in cf.locations:
                        cf = self._create_conversion_cf(ex_cf, rq, cx, qq.origin, loc, cf[loc], overwrite=False)
                        logging.warning('Created unit-conversion CF\n%s' % cf)
                else:
                    logging.warning('FactorCollision for %s into %s: %g != %g' % (fb, cx, cf.value, ex_cf.value))
                    self._dupes.append(DuplicateCharacterization(qq.link, fb, cx.as_list(), qq.origin, cf.flowable, None))

        self._factors_for_later[quantity] = True
        print('Imported %d factors for %s' % (count, quantity))

    def _check_factors(self, qq):
        if hasattr(self._factors_for_later[qq], 'factors'):
            self.import_cfs(self._factors_for_later.pop(qq))

    def _find_exact_cf(self, qq, fb, cx, origin):
        try:
            ql = self._qlookup(qq, fb)
        except NoFQEntry:
            return None
        # cfs = ql._context_origin(cx, origin=origin)
        # if fb is self._fm['water']:  # this is not how to handle this
        #     cx = self.add_subcontext(cx, 'flow', flowable)
        cfs = ql.find(cx, dist=0, origin=origin)  # sliiiiightly slower but much better readability
        if len(cfs) > 0:
            if len(cfs) > 1:  # can only happen if qq's origin is None ??
                try:
                    return next(cf for cf in cfs if cf.origin is None)
                except StopIteration:
                    # this should never happen
                    return next(cf for cf in cfs if cf.origin == qq.origin)
            return list(cfs)[0]
        return None

    @staticmethod
    def _store_cf(cl, context, new_cf):
        """
        Assigns the cf to the mapping; does subclass-specific collision checking
        :param cl:
        :param new_cf:
        :return:
        """
        try:
            cl.add(new_cf, key=context)
        except TypeError:
            print(type(cl))
            raise

    def _qassign(self, qq, fb, new_cf, context=None):
        """

        :param qq:
        :param fb:
        :param new_cf:
        :param context:
        :return:
        """
        ''' # this is done in _find_exact_cf already
        # don't characterize "Water" per se-- give it a subcontext
        if context is not NullContext and fb is self._fm['Water']:
            context = self.add_subcontext(context, 'flow', new_cf.flowable)
        '''

        super(LciaEngine, self)._qassign(qq, fb, new_cf, context)
        self._origins.add(new_cf.origin)

    def _check_merge_conflicts(self, dominant, *syns):
        """
        This just brute forces it which must be ungodly slow, and btw it also hasn't been tested

        Inputs must all be canonical _fm entries

        :param dominant:
        :param syns:
        :return:
        """
        fq_conflicts = []
        for f_dict in self._q_dict.values():
            try:
                combo = f_dict[dominant]
            except KeyError:
                combo = self._cl_typ()
            for syn in syns:
                if syn not in f_dict:
                    continue
                cl = f_dict[syn]
                for k in cl.keys():  # of course, we have to check all contexts
                    if k in combo:
                        cand = ({c.value for c in combo[k]},
                                {c.value for c in cl[k]})
                        if cand[0] == cand[1]:
                            continue
                        fq_conflicts.append(cand)  # but it's really only a conflict if the cf values differ

        if len(fq_conflicts) > 0:
            print('%d Merge conflicts encountered' % len(fq_conflicts))
            raise FactorConflict(fq_conflicts)

    def merge_flowables(self, dominant, *syns):
        """
        Combine together a list of synonyms under a dominant term.  If the synonyms are existing flowables, they are
        merged with the dominant term; otherwise, they are added as synonyms.
        :param dominant: must be known to the flowables database
        :param syns: 0 or more terms or flowables
        :return:
        """
        dom_fb = self.get_flowable(dominant)
        to_merge = set()
        for syn in syns:
            try:
                c = self.get_flowable(syn)
                if c is dom_fb:
                    continue
                to_merge.add(c)
            except KeyError:
                self.add_synonym(dom_fb, syn)
        if to_merge:
            return self._merge_terms(dom_fb, *to_merge)
        else:
            return dom_fb

    def save_flowables(self, filename=None):
        self._fm.save(filename)

    def save_contexts(self, filename=None):
        self._cm.save(filename)

    def flowables(self, search=None, origin=None, new=False, **kwargs):
        """
        Adds ability to filter by origin-- note this is exclusive to the ability to filter by quantity
        :param search:
        :param origin:
        :param new: [False] if True, only return flowables that were not known originally
        :return: generates flowable strings
        """
        if origin is None:
            _iter = super(LciaEngine, self).flowables(search=search, **kwargs)
        else:
            if search is None:
                _iter = self._fb_by_origin[origin]
            else:
                _iter = (str(x) for x in self._fm.objects_with_string(search) if x in self._fb_by_origin[origin])

        for k in _iter:
            if new:
                if k in self._fb_by_origin[None]:
                    continue
            yield k

    def _factors_for_flowable(self, fb, qq, cx, **kwargs):
        """
        detach lookup for cleanness. canonical everything
        :param fb:
        :param qq:
        :param cx:
        :param kwargs: used in subclasses
        :return:
        """
        self._check_factors(qq)
        try:
            cl = self._qlookup(qq, fb)
        except NoFQEntry:
            return
        if cx is None:
            for v in cl.cfs():
                yield v
        else:
            for v in cl.find(cx, **kwargs):
                yield v

    def factors_for_quantity(self, quantity, flowable=None, context=None, **kwargs):
        self._check_factors(self._canonical_q(quantity))
        return super(LciaEngine, self).factors_for_quantity(quantity, flowable=flowable, context=context, **kwargs)

    def _serialize_qdict(self, quantity, values=False, origin=None):
        _ql = self._qaccess(quantity)
        d = {}
        for fb, cl in _ql.items():
            _od = cl.serialize_for_origin(origin, values=values)
            if len(_od) > 0:
                d[str(fb)] = _od
        return d



from ..lcia_results import LciaResult
from ..exchanges import ExchangeValue
from ..contexts import NullContext
from antelope import QuantityRequired, ConversionReferenceMismatch, BackgroundRequired, EntityNotFound, comp_dir


UNRESOLVED_ANCHOR_TYPE = 'term'  # this is used when an anchor node's origin cannot be resolved


class FlowConversionError(Exception):
    pass


class UnresolvedAnchor(Exception):
    pass


class UnCachedScore(Exception):
    """
    means that we have an LCIA-only node whose score has not been set for the requested LCIA method
    """
    pass


class Anchor(object):
    """
    An anchor is the distal partner to an exchange.
    """
    @classmethod
    def null(cls, parent):
        return cls(parent, None)

    _node = None
    _anchor_flow = None

    def __init__(self, parent, anchor_node, anchor_flow=None, descend=None):
        """
        An anchor can be one of the following five types:
         - null: a cut-off. The fragment is an input/output of the spanner
         - foreground: the fragment is its own anchor.
         - process: the fragment is anchored to a process with inventory
         - sub-fragment: the fragment is anchored to another spanner
         - context: the fragment is an exchange with a context

        :param parent:
        :param anchor_node:
        :param anchor_flow:
        :param descend:
        """
        self._parent = parent
        self._node = anchor_node
        self.anchor_flow = anchor_flow
        self._descend = descend
        self._score_cache = dict()

    @property
    def is_null(self):
        return self._node is None

    @property
    def is_fg(self):
        return (not self.is_null) and (self.node is self._parent)

    @property
    def is_process(self):
        return (not self.is_null) and (self.node.entity_type == 'process')

    @property
    def is_frag(self):
        """
        Termination is a fragment
        :return:
        """
        return (not self.is_null) and (self.node.entity_type == 'fragment')

    @property
    def is_subfrag(self):
        """
        Termination is a non-self fragment.

        :return:
        """
        return self.is_frag and not self.is_fg

    @property
    def is_context(self):
        """
        termination is a context
        :return:
        """
        return (not self.is_null) and (self._node.entity_type == 'context')

    @property
    def is_emission(self):
        """
        Pending context refactor
        :return:
        """
        return self.is_context and self.node.elementary

    @property
    def node(self):
        return self._node

    @property
    def name(self):
        if self.is_null:
            name = self.anchor_flow['Name']
        elif self.is_context:
            name = '%s, %s' % (self.anchor_flow['Name'], self.node.name)
        else:
            name = self.node.name
        return name

    @property
    def unit(self):
        return self.anchor_flow.unit

    def __str__(self):
        """
        This is repeated at least once in the Anchor model
        :return:
          '---:' = fragment I/O
          '-O  ' = foreground node
          '-*  ' = process
          '-#  ' - sub-fragment (aggregate)
          '-#::' - sub-fragment (descend)
          '-B ' - terminated background
          '--C ' - cut-off background
          '--? ' - ungrounded catalog ref
        """
        if self.is_null:
            term = '---:'  # fragment IO
        elif self.is_fg:
            term = '-O  '
        elif self.is_context:
            if self.is_emission:
                term = '-== '
            elif self.node is NullContext:
                term = '-)  '
            else:
                # TODO: intermediate contexts don't present as cutoffs (because is_null is False)
                term = '-cx '
        elif self.node.entity_type == 'process':
            if self._parent.is_background:
                term = '-B* '
            else:
                term = '-*  '
        elif self.node.entity_type == 'fragment':
            if self.descend:
                term = '-#::'
            else:
                term = '-#  '
        elif self.node.entity_type == UNRESOLVED_ANCHOR_TYPE:
            term = '--? '
        else:
            raise TypeError('I Do not understand this term for frag %.7s' % self._parent.uuid)
        return term

    @property
    def flow_conversion(self):
        """
        express the parent's flow in terms of the quantity of the term flow.
        There are two ways to do this, each case involving the quantity relation on either the parent flow or the
        term flow, between the two quantities (parent flow's r.q. is the reference quantity; term flow's r.q. is the
        query quantity).

        In each case, we want the flow's native term manager to perform the conversion using ITS OWN canonical
        quantities.  The assumption is that the parent flow's r.q. is tied to our local LciaEngine, while the
        term flow's r.q. could be either local or remote.

        The QuantityRef.quantity_relation() implicitly assumes that the invoking quantity is the QUERY quantity, so
        the "forward" (natural parent->node) direction uses the remote flow's r.q. - but we do the "reverse" direction
        first because it's local.

        how to deal with scenario cfs? tbd
        problem is, the term doesn't know its own scenario

        :return: float = amount in term_flow ref qty that corresponds to a unit of fragment flow's ref qty
        """
        '''
        What are all the possibilities?
         the parent quantity knows a characterization for the parent flow w.r.t. the term quantity
         the parent quantity knows a characterization for the term flow w.r.t. the term quantity
         the term quantity knows a characterization for the parent flow w.r.t. the parent quantity
         the term quantity knows a characterization for the term flow w.r.t. the parent quantity

        "hey look at me, net calorific value. see that mass flow f4- a unit of it is worth 35 of me"
        is how the database is constructed.

        but each call to quantity_relation should check for both forward and reverse matches
        '''
        if self.anchor_flow.reference_entity == self._parent.flow.reference_entity:
            return 1.0
        parent_q = self._parent.flow.reference_entity
        term_q = self.anchor_flow.reference_entity
        if term_q is None:
            return 0.0

        # first - natural - ask our parent flow if fit can convert to term quantity
        try:
            rev = self._parent.flow.cf(term_q, context=self.node.external_ref)
        except (QuantityRequired, NotImplementedError, ConversionReferenceMismatch):
            rev = None

        if rev:
            return rev

        # then ask the term_flow if it can convert to parent quantity
        try:
            fwd = self.anchor_flow.cf(parent_q)
        except (QuantityRequired, NotImplementedError, ConversionReferenceMismatch):
            fwd = None

        if fwd:
            return 1.0 / fwd

        # then ask if our parent qty can convert term_flow
        try:
            rev_c = parent_q.quantity_relation(self.anchor_flow, term_q, context=(self.node.origin,
                                                                                  self.node.external_ref))
        except (QuantityRequired, NotImplementedError):
            rev_c = None
        except ConversionReferenceMismatch:
            try:
                rev_c = parent_q.quantity_relation(self.anchor_flow, term_q, context=None)
            except ConversionReferenceMismatch:
                rev_c = None

        if rev_c:
            return 1.0 / rev_c.value

        # last, ask if remote quantity recognizes *our* flow
        print(' %s: flow_conversion: untested' % self._parent.link)
        try:
            fwd_c = term_q.quantity_relation(self._parent.flow, parent_q, context=(self.node.origin,
                                                                                   self.node.external_ref))
        except (QuantityRequired, NotImplementedError, ConversionReferenceMismatch):
            fwd_c = None

        if fwd_c:
            print('reverse q-hit')
            return fwd_c.value
        print(' %s: flow_conversion FAILED' % self._parent.link)
        raise FlowConversionError('Zero CF found relating %s to %s' % (self.anchor_flow, self._parent.flow))

    @property
    def inbound_exchange_value(self):
        if self.is_process:
            if self._anchor_flow.direction == self._parent.direction:
                return -1.0
            return 1.0

    @property
    def node_weight_multiplier(self):
        return self.flow_conversion * self.inbound_exchange_value

    @property
    def anchor_flow(self):
        if self.is_process:
            return self._anchor_flow.flow
        else:
            return self._anchor_flow

    @anchor_flow.setter
    def anchor_flow(self, flow):
        if self.is_process:
            self._anchor_flow = self.node.reference(flow)
        elif flow is None:
            if self.is_frag:
                self._anchor_flow = self.node.flow
            else:
                self._anchor_flow = self._parent.flow
        else:
            self._anchor_flow = flow

    @property
    def direction(self):
        if self.is_process:
            return self._anchor_flow.direction
        return comp_dir(self._parent.direction)

    @property
    def descend(self):
        if self.is_null:
            return None
        return bool(self._descend)

    @descend.setter
    def descend(self, value):
        if self.is_null:
            raise TypeError('may not set descend on a null anchor')
        else:
            self._descend = value

    def add_lcia_score(self, quantity, score, scenario=None):
        res = LciaResult(quantity, scenario=scenario)
        res.add_summary(self._parent.external_ref, self._parent, 1.0, score)
        self._score_cache[quantity] = res

    def unobserved_exchanges(self):
        """
        Generator which yields exchanges from the term node's inventory that are not found among the child flows, for
          LCIA purposes

        Challenge here going forward: we made some kind of normative decision early on that terminations do not know
        their own scenarios, that the fragment maps scenario to termination. The problem is that now terminations
        cannot themselves carry out traversal on the term_node because they don't know what scenarios to pass.

        The upshot is that we cannot meaningfully compute "unobserved exchanges" for subfragments, since we don't
        know our scenario.

        :return:
        """
        if self.is_context:
            x = ExchangeValue(self._parent, self.anchor_flow, self._parent.direction, termination=self.node,
                              value=self.node_weight_multiplier)  # TODO: need to figure out how we are handling locales
            yield x
        elif self.is_frag:  # we need to send the fragment's cutoffs-- but that cannot occur here
            raise NotImplementedError
        else:
            if self._parent.is_background:  # or len(list(self._parent.child_flows)) == 0:
                # ok we're bringing it back but only because it is efficient to cache lci
                for x in self.node.lci(ref_flow=self.anchor_flow):
                    yield x
            else:
                for x in self.node.unobserved_lci(self._parent.child_flows, ref_flow=self.anchor_flow):
                    yield x  # this should forward out any cutoff exchanges

    def compute_unit_score(self, quantity_ref, refresh=False, **kwargs):
        """
        four different ways to do this.
        0- we are a subfragment-- no direct impacts unless non-descend, which is caught earlier
        1- parent is bg: ask catalog to give us bg_lcia (process or fragment)
        2- get fg lcia for unobserved exchanges

        If
        :param quantity_ref:
        :param refresh:
        :return:
        """
        try:
            locale = self.node['SpatialScope']
        except KeyError:
            locale = 'GLO'

        try:
            res = self.node.bg_lcia(quantity_ref, observed=self._parent.child_flows, ref_flow=self.anchor_flow,
                                    locale=locale, **kwargs)
        except (BackgroundRequired, EntityNotFound, NotImplementedError):
            try:
                res = quantity_ref.do_lcia(self.unobserved_exchanges(), locale=locale,
                                           refresh=refresh, **kwargs)
            except (QuantityRequired, NotImplementedError):
                raise UnresolvedAnchor

        if isinstance(res, list):
            [k.scale_result(self.inbound_exchange_value) for k in res]
        else:
            res.scale_result(self.inbound_exchange_value)
        return res

    def score_cache(self, quantity=None, refresh=False, **kwargs):
        """
        only process-terminations are cached
        remote fragments that come back via the API can have cached scores as well, but local subfragments should not
        get cached.

        :param quantity:
        :param refresh: If True, re-compute unit score even if it is already present in the cache.
        :param kwargs:
        :return:
        """
        if quantity is None:
            return self._score_cache

        if self.is_null or self.is_fg:
            return LciaResult(quantity)

        if refresh:
            self._score_cache.pop(quantity, None)

        if quantity in self._score_cache:
            return self._score_cache[quantity]
        else:
            if self.is_frag:  # but not fg, ergo subfrag
                raise UnCachedScore(quantity)
            else:
                res = self.compute_unit_score(quantity, **kwargs)
            if isinstance(res, list):
                for k in res:
                    self._score_cache[k.quantity] = k
            self._score_cache[quantity] = res
            return self._score_cache[quantity]

    def clear_score_cache(self, quantity=None):
        if quantity is not None:
            self._score_cache.pop(quantity, None)
        else:
            self._score_cache = dict()



from antelope import EntityNotFound, NoAccessToEntity, NullEntity, ItemNotFound
from ..contexts import NullContext


class BasicImplementation(object):
    def __init__(self, archive, **kwargs):
        """
        Provides common features for an interface implementation: namely, an archive and a privacy setting. Also
        provides access to certain common methods of the archive.  This should be the base class for interface-specific
        implementations.

        Requires an archive with the following attributes:
         - ref - report semantic reference
         - source - report physical data source
         - static - a boolean indicating whether the full contents of the archive are loaded into memory
         - get_uuid() - deprecated - only present for compatibility reasons
         - __getitem__ - retrieve already-loaded entity
         - retrieve_or_fetch_entity() - _fetch abstract method must be implemented

        All of these requirements are met by the standard ArchiveImplementation, with the exception of the _fetch
        abstract method.

        Since a recent change that removed the 'basic' interface as a default for all resources, this must be
        explicitly assigned to at least one resource in order for a query to be valid.  The basic interface should
        be assigned to the resource that meets the following requirements:
         - most comprehensive source of information about entity properties (e.g. documentary pseudo-interface)
         - easiest to load (e.g. a non-static)

        :param archive: an LcArchive
        :param privacy: No longer used. Privacy is enforced at the server and not the resource (where it was phony
        from the beginning)
        """
        self._archive = archive

    def validate(self):
        """
        way to check that a query implementation is valid without querying anything
        :return:
        """
        return self.origin

    def is_me(self, archive):
        return self._archive is archive

    @property
    def origin(self):
        return self._archive.ref

    def __str__(self):
        return '%s for %s (%s)' % (self.__class__.__name__, self.origin, self._archive.source)

    def __getitem__(self, item):
        return self._archive[item]

    def _dereference_entity(self, external_ref):
        """
        returns a real local entity (non ref) for which is_entity is True
        :param external_ref:
        :return:
        """
        if hasattr(external_ref, 'external_ref'):
            eref = external_ref
            external_ref = eref.external_ref
        else:
            eref = None
        entity = self.get(external_ref)
        # if entity:
        if not entity.is_entity:
            raise NoAccessToEntity(self.origin, entity.link)
        return entity

    def get_item(self, external_ref, item):
        """
        In this, we accept either an external_ref or an entity reference itself.  If the latter, we dereference via
        the archive to an actual entity, which we then ask for the item.  If the dereference and the reference are the
        same, throws an error.
        :param external_ref:
        :param item:
        :return:
        """
        entity = self._dereference_entity(external_ref)
        if entity.has_property(item):
            obj = entity[item]
            if obj is None:
                raise ItemNotFound
            return obj
        raise ItemNotFound('%s: %s [%s]' % (self.origin, external_ref, item))
        # raise EntityNotFound(external_ref)

    def get_reference(self, key):
        entity = self._dereference_entity(key)
        if entity is None:
            return None
        if entity.entity_type == 'process':
            # need to get actual references with exchange values-- not the reference_entity
            return [x for x in entity.references()]
        if entity.reference_entity is None:
            return NullEntity(self.origin)
        return entity.reference_entity

    def get_uuid(self, external_ref):
        u = self._fetch(external_ref)
        if u is None:
            return False
        if u.uuid is None:
            return False
        return u.uuid

    def _fetch(self, external_ref, **kwargs):
        if external_ref is None:
            return None
        if self._archive.static:
            return self._archive[external_ref]
        try:
            return self._archive.retrieve_or_fetch_entity(external_ref, **kwargs)
        except (KeyError, NotImplementedError, IndexError):
            return None

    '''
    def lookup(self, external_ref, **kwargs):
        if self._fetch(external_ref, **kwargs) is not None:
            return True
        return False
    '''

    def get(self, external_ref, **kwargs):
        """
        :param external_ref: may also be link, as long as requested origin is equal or lesser in specificity
        :param kwargs:
        :return: entity or None
        """
        if external_ref is None:
            raise EntityNotFound(None)
        e = self._fetch(external_ref, **kwargs)
        if e is not None:
            return e
        if isinstance(external_ref, int):  # we got away with this before by falling back on NSUUIDs
            external_ref = str(external_ref)
        e = self._fetch(external_ref, **kwargs)
        if e is not None:
            return e
        er_s = external_ref.split('/')
        if self.origin.startswith(er_s[0]):
            e = self._fetch('/'.join(er_s[1:]), **kwargs)
            if e is not None:
                return e
        raise EntityNotFound(external_ref)

    def get_context(self, term, **kwargs):
        """
        I think this needs to be moved into the quantity interface
        :param term:
        :param kwargs:
        :return:
        """
        cx = self._archive.tm[term]
        if cx is None:
            return NullContext
        if cx.fullname == cx.name:
            cx.add_origin(self.origin)
        return cx

    def is_lcia_engine(self, **kwargs):
        """
        suggests expansion to a graph-based TM
        :param kwargs:
        :return:
        """
        if hasattr(self._archive, 'tm'):
            return self._archive.tm.is_lcia_engine
        return False

    def synonyms(self, item, **kwargs):
        if hasattr(self._archive, 'tm'):
            return self._archive.tm.synonyms(item)
        # yield from ()

    def properties(self, external_ref, **kwargs):
        e = self._dereference_entity(external_ref)
        for i in e.properties():
            yield i

    def bg_lcia(self, process, query_qty, ref_flow=None, *kwargs):
        """
        This needs to be handled by a query with lci() access, or by a subclass
        :param process:
        :param query_qty:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        raise NotImplementedError



"""
The CLookup is a classic example of trying to solve a very complex data problem with a very complex data structure.
It does not entirely succeed, but it works well enough as a stopgap until I have a chance to give an honest crack at
a graph database.

Flow characterization is the very complex data problem: the "quantity relation" maps a large, diverse set of inputs:
 - flowable (substance)
 - reference quantity (in our data system, tied to flowable)
 - query quantity
 - context / compartment
 - location / locale
to a numeric output, namely the amount of the query quantity that corresponds to a unit of the reference quantity.
Technically, this amount also has uncertainty / other quantitative characteristics.

The idea behind a CLookup is that it contains all known characterizations for a given flowable [substance] with respect
to a given quantity.  The CLookup is selected by specifying the flowable and the quantity, and then the CLookup is used
to retrieve a set of Characterization objects for a given context (hence the 'C' in 'CLookup'). The characterization
already stores a mapping of locale to factor, and also stores the flowable (with its native reference quantity used
to interpret the factor).

So it solves a very narrow portion of the problem and leaves a lot to outside code.

The dream would be to design a graph database that held all of these parameters and magically obtained all the factors
that applied to a given query-- that graph database would replace the current Term Manager and everything else under
its hood. But first we will learn to walk...
"""
from ..contexts import Context, NullContext
from ..characterizations import DuplicateCharacterizationError


class QuantityMismatch(Exception):
    pass


class DuplicateOrigin(Exception):
    pass


class CLookup(object):
    """
    A CLookup is a kind of fuzzy dictionary that maps context to best-available characterization. A given CLookup is
    associated with a single quantity and a specific flowable.  The query then provides the compartment and returns
    either: a set of best-available characterizations; a single characterization according to a selection rule; or a
    characterization factor (float) depending on which method is used.
    """
    def __init__(self):
        self._dict = dict()
        self._q = None

    def __repr__(self):
        return '%s(%s: %d contexts)' % (self.__class__.__name__, self._q, len(self._dict))

    def __getitem__(self, item):
        """
        Returns
        :param item:
        :return:
        """
        if item is None:
            item = NullContext
        if not isinstance(item, Context):
            raise TypeError('Supplied CLookup key is not a Context: %s (%s)' % (item, type(item)))
        if item in self._dict:
            return self._dict[item]
        return []

    def __contains__(self, item):
        return item in self._dict

    def _check_qty(self, cf):
        if self._q is None:
            self._q = cf.quantity
        else:
            if cf.quantity != self._q:
                raise QuantityMismatch('Inbound: %s\nCurrent: %s' % (cf.quantity, self._q))

    def add(self, value, key=None):
        if key is None:
            key = value.context
        if isinstance(key, Context):
            self._check_qty(value)
            if key not in self._dict:
                self._dict[key] = []
            '''
            if any(k.origin == value.origin for k in self._dict[key]):
                raise DuplicateOrigin(value.origin)
            '''
            if value in self._dict[key]:
                try:
                    exis = next(v for v in self._dict[key] if hash(v) == hash(value))
                    for l in value.locations:
                        exis[l] = value[l]
                except DuplicateCharacterizationError:
                    self._dict[key].append(value)
            else:
                self._dict[key].append(value)
        else:
            raise ValueError('Context is not valid: %s (%s)' % (key, type(key)))

    def remove(self, value):
        key = value.context
        self._dict[key].remove(value)

    def keys(self):
        return self._dict.keys()

    def cfs(self):
        for c, cfs in self._dict.items():
            for cf in cfs:
                yield cf

    def _context_origin(self, item, origin):
        if origin is None:
            return list(self.__getitem__(item))
        else:
            return [k for k in self.__getitem__(item) if k.origin == origin]

    def find(self, item, dist=1, return_first=True, origin=None):
        """
        Hunt for a matching compartment. 'dist' param controls the depth of search:
          dist = 0: equivalent to __getitem__
          dist = 1: also check compartment's children (subcompartments), to any depth, returning all CFs encountered
            (unless return_first is True, in which case all CFs from the first nonempty compartment found are returned)
          dist = 2: also check compartment's parents (excluding root=Null context)
          dist = 3: also check all compartment's parents until root. Useful for finding unit conversions.
        By default (dist==1), checks compartment self and children. Returns a set.
        :param item: a Compartment
        :param dist: how far to search (with limits) (default: 1= compartment + children)
        :param return_first: stop hunting as soon as a cf is found
        :param origin: [None] if present, only return cfs whose origins match the specification
        :return: a list of characterization factors that meet the query criteria, ordered by increasing dist
        """
        #!TODO: This should be performed by the term manager, not the CLookup (though that is the whole point of CLookup)
        if not isinstance(item, Context):
            return []

        def found(res):
            return len(res) > 0 and return_first
        results = self._context_origin(item, origin)
        if found(results):
            return results

        if dist > 0:
            for s in item.self_and_subcompartments:  # note: this is depth first
                if s is item:
                    continue  # skip self, just recurse subcompartments
                results += self._context_origin(s, origin)
                if found(results):
                    return results

        if dist > 1:
            while item.parent is not None:
                item = item.parent
                results += self._context_origin(item, origin)
                if found(results):
                    return results

        if dist > 2:
            while item is not None:
                item = item.parent
                results += self._context_origin(item, origin)
                if found(results):
                    return results

        return results

    def find_first(self, item, dist=3):
        cfs = self.find(item, dist=dist, return_first=True)
        return list(cfs)[0]

    @staticmethod
    def _ser_set(cf_set, values=False):
        return {cf.origin: cf.serialize(values=values, concise=True) for cf in cf_set}

    def serialize(self, values=False):
        return {str(c): self._ser_set(cfs, values=values) for c, cfs in self._dict.items()}

    def serialize_for_origin(self, origin, values=False):
        """
        Note: In the event that the CLookup includes multiple CFs for the same flowable and the same origin, only the
        first (at random) will be included, because originally I had disallowed multiple CFs for the same origin.
        :param origin:
        :param values:
        :return:
        """
        cxs = dict()
        for c, cfs in self._dict.items():
            if origin:
                try:
                    cf_filt = next(cf for cf in cfs if cf.origin == origin)  # duplicate entries per origin not allowed
                except StopIteration:
                    continue
            else:
                if len(cfs) == 0:
                    continue
                else:
                    cf_filt = next(cf for cf in cfs)

            cxs[str(c)] = cf_filt.serialize(values=values, concise=True)
        return cxs


class FactorCollision(Exception):
    pass


class SCLookup(CLookup):
    """
    A Strict CLookup that permits only one CF to be stored per compartment and raises an error if an additional one
    is added.
    """
    def add(self, value, key=None):
        if key is None:
            key = value.context
        if key in self._dict and len(self._dict[key]) > 0:
            existing = self._dict[key][0]
            if existing.value == value.value:
                return
            print('Collision with context: %s' % repr(key))
            print(repr(value))
            print('%s current' % repr(existing))
            raise FactorCollision('This context already has a CF defined!')
        super(SCLookup, self).add(value, key)



import unittest
from ...entities import LcFlow, LcQuantity
from ..lcia_engine import LciaEngine


class TestBiogenicCo2(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.mass = LcQuantity.new('mass', 'kg', origin='test')

    def test_co2_name(self):
        f = LcFlow.new('carbon dioxide', ref_qty=self.mass)
        self.assertTrue(f.is_co2)
        self.assertFalse(f.quell_co2)

    def test_co2_spec(self):
        f = LcFlow.new('co2', ref_qty=self.mass, is_co2=True)
        self.assertTrue(f.is_co2)
        self.assertFalse(f.quell_co2)

    def test_co2_cas(self):
        f = LcFlow.new('carbondioxide', ref_qty=self.mass, casnumber='124389')
        self.assertTrue(f.is_co2)
        self.assertFalse(f.quell_co2)

    def test_co2_catch_cas(self):
        f = LcFlow.new('carbon 2-oxide', ref_qty=self.mass)
        self.assertFalse(f.is_co2)
        f['casnumber'] = '124389'
        self.assertTrue(f.is_co2)
        self.assertFalse(f.quell_co2)

    def test_co2_biogenic(self):
        f = LcFlow.new('co2, in air', ref_qty=self.mass)
        self.assertFalse(f.is_co2)
        self.assertFalse(f.quell_co2)
        f.is_co2 = True
        self.assertTrue(f.is_co2)
        self.assertTrue(f.quell_co2)

    def test_co2_lcia_engine(self):
        l = LciaEngine()
        co2 = LcFlow.new('co2, biotic', ref_qty=self.mass, synonyms='carbon dioxide', origin='test',
                         context='in air')
        self.assertFalse(co2.is_co2)
        l.add_flow(co2)
        self.assertTrue(co2.is_co2)
        self.assertTrue(co2.quell_co2)


if __name__ == '__main__':
    unittest.main()



import re
from itertools import chain

from .basic import BasicImplementation
from antelope import BackgroundInterface, EntityNotFound, comp_dir  # , ProductFlow
from antelope.models import ExteriorFlow
from antelope_core.contexts import Context


class NonStaticBackground(Exception):
    pass


def search_skip(entity, search):
    if search is None:
        return False
    return not bool(re.search(search, str(entity), flags=re.IGNORECASE))


class BackgroundImplementation(BasicImplementation, BackgroundInterface):
    """
    The default Background Implementation exposes an ordinary inventory database as a collection of LCI results.
    Because it does not perform any ordering, there is no way to distinguish between foreground and background
    elements in a database using the proxy. It is thus inconsistent for the same resource to implement both
    inventory and [proxy] background interfaces.
    """
    def __init__(self, *args, **kwargs):
        super(BackgroundImplementation, self).__init__(*args, **kwargs)

        self._index = None

    def check_bg(self, **kwargs):
        self.setup_bm(**kwargs)
        return bool(self._index is not None)

    def setup_bm(self, index=None):
        """
        Requires an index interface or catalog query <-- preferred
        This must provide .get() and .get_context() (so really it should maybe be 'basic'
        The trivial implementation uses .flows() and .targets() to mock up an exterior flows method
        :param index:
        :return:
        """
        if self._index is None:
            if index is None:
                # print('%%%%%% Setting up Background Impl for %s from archive %s' % (self.origin, self._archive))
                self._index = self._archive.make_interface('index')
            else:
                # print('%%%%%% Setting up Background Impl for %s from index %s' % (self.origin, index))
                self._index = index

    def _ensure_ref_flow(self, ref_flow):
        if ref_flow is not None:
            if isinstance(ref_flow, str) or isinstance(ref_flow, int):
                ref_flow = self._archive.retrieve_or_fetch_entity(ref_flow)
        return ref_flow

    def foreground_flows(self, search=None, **kwargs):
        """
        No foreground flows in the proxy background
        :param search:
        :param kwargs:
        :return:
        """
        for i in []:
            yield i

    def background_flows(self, search=None, **kwargs):
        """
        all process reference flows are background flows
        :param search:
        :param kwargs:
        :return:
        """
        self.check_bg()
        for p in self._archive.entities_by_type('process'):
            for rx in p.references():
                if search_skip(p, search):
                    continue
                yield rx  # ProductFlow(self._archive.ref, rx.flow, rx.direction, p, None) don't need this

    def exterior_flows(self, direction=None, search=None, **kwargs):
        """
        Exterior flows are all flows that do not have interior terminations (i.e. not found in the index targets)
        Elementary contexts have a sense, but intermediate contexts do not [necessarily]- so we need some way to
        determine their directionality.  This whole implementation is just a stand-in anyway- the important thing
        is that this is handled correctly in tarjan
        :param direction:
        :param search:
        :param kwargs:
        :return:
        """
        self.check_bg()
        for f in self._index.flows():
            if search_skip(f, search):
                continue
            try:
                next(self._index.targets(f.external_ref, direction=direction))
            except StopIteration:
                cx = self._index.get_context(f.context)
                dirn = comp_dir(cx.sense)  # this is already w.r.t. interior
                '''
                if self.is_elementary(f):
                    yield ExteriorFlow(self._archive.ref, f, 'Output', f['Compartment'])
                else:
                    yield ExteriorFlow(self._archive.ref, f, 'Output', None)
                '''
                yield ExteriorFlow.from_background(f, dirn, cx)

    def consumers(self, process, ref_flow=None, **kwargs):
        """
        Not supported for trivial backgrounds
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in []:
            yield i

    def dependencies(self, process, ref_flow=None, **kwargs):
        """
        All processes are LCI, so they have no dependencies
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in []:
            yield i

    def emissions(self, process, ref_flow=None, **kwargs):
        """
        All processes are LCI, so they have only exterior flows. Emissions are the exterior flows with elementary
        contexts
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in self.lci(process, ref_flow=ref_flow, **kwargs):
            if isinstance(i.termination, Context):
                if i.termination.elementary:
                    yield i

    def cutoffs(self, process, ref_flow=None, **kwargs):
        """
        All processes are LCI, so they have only exterior flows. Emissions are the exterior flows with non-elementary
        contexts
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in self.lci(process, ref_flow=ref_flow, **kwargs):
            if isinstance(i.termination, Context):
                if i.termination.elementary:
                    continue
            yield i

    def foreground(self, process, ref_flow=None, **kwargs):
        self.check_bg()
        ref_flow = self._ensure_ref_flow(ref_flow)
        p = self._index.get(process)
        yield p.reference(ref_flow)  # should be just one exchange

    def is_in_scc(self, process, ref_flow=None, **kwargs):
        """
        Distinction between is_in_background and is_in_scc will reveal the proxy nature of the interface
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        return False  # proxy has no knowledge of SCCs

    def is_in_background(self, process, ref_flow=None, **kwargs):
        self.check_bg()
        try:
            self._archive.retrieve_or_fetch_entity(process)
        except EntityNotFound:
            return False
        return True

    def ad(self, process, ref_flow=None, **kwargs):
        for i in []:
            yield i

    def bf(self, process, ref_flow=None, **kwargs):
        for i in []:
            yield i

    def lci(self, process, ref_flow=None, **kwargs):
        self.check_bg()
        ref_flow = self._ensure_ref_flow(ref_flow)
        p = self._archive.retrieve_or_fetch_entity(process)
        for x in p.inventory(ref_flow=ref_flow):
            yield x

    def sys_lci(self, demand, **kwargs):
        """
        For LCI, we simply yield process direct exchanges as LCI.
        For sys_lci, we should just do the same. yield the supplied demand as a degenerate LCI.
        :param demand:
        :param kwargs:
        :return:
        """
        for y in demand:
            yield y

    def sys_lcia(self, process, query_qty, observed=None, ref_flow=None, **kwargs):
        """
        returns an LciaResult object, aggregated as appropriate depending on the interface's privacy level.
        This is an ensemble function that stitches together bg functions with quantity access.

        :param process:
        :param query_qty: must be an operable quantity_ref. the process must have exchange access
        :param observed: iterable of DirectedFlows (flow: FlowSpec, direction: str)
        :param ref_flow:
        :param kwargs:
        :return:
        """
        p_ref = self.get(process)  # a ref-- unless this was a basic impl. hrm.
        if p_ref.is_entity:
            raise NotImplementedError  # we can't proceed
        if observed is None:
            observed = ()
        lci = p_ref.unobserved_lci(observed, ref_flow=ref_flow)
        '''
        obs = set()
        for k in observed:
            if k.value < 0:
                obs.add((k.flow.external_ref, comp_dir(k.direction)))
            else:
                obs.add((k.flow.external_ref, k.direction))
        # obs = set((k.flow.external_ref, k.direction) for k in observed)
        if len(obs) > 0:
            exts = chain(p_ref.emissions(ref_flow=ref_flow),
                         p_ref.cutoffs(ref_flow=ref_flow))
            incl = (k for k in p_ref.dependencies(ref_flow=ref_flow) if (k.flow.external_ref, k.direction) not in obs)
            ext = (k for k in exts if (k.flow.external_ref, k.direction) not in obs)
            lci = chain(self.sys_lci(incl), ext)
        else:
            lci = p_ref.lci(ref_flow=ref_flow)
        '''  # aggregation
        return query_qty.do_lcia(lci, **kwargs)



OLCA_MAPPING = {
    'Currency':	{'referenceCurrency': 'refCurrency'},
    'Exchange': {'avoidedProduct': 'isAvoidedProduct',
                 'input': 'isInput',
                 'quantitativeReference': 'isQuantitativeReference'},
    'Flow': {'infrastructureFlow': 'isInfrastructureFlow'},
    'FlowPropertyFactor': {'referenceFlowProperty': 'isRefFlowProperty'},
    'ImpactCategory': {'referenceUnitName': 'refUnit'},
    'Parameter': {'inputParameter': 'isInputParameter'},
    'Process': {'infrastructureProcess': 'isInfrastructureProcess'},
    'ProcessDocumentation': {'copyright': 'isCopyrightProtected'},
    'ProductSystem': {'referenceExchange': 'refExchange',
                      'referenceProcess': 'refProcess'},
    'Unit': {'referenceUnit': 'isRefUnit'}
}
